# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Audio summaries and TensorFlow operations to create them, V2 versions.

An audio summary stores a rank-2 string tensor of shape `[k, 2]`, where
`k` is the number of audio clips recorded in the summary. Each row of
the tensor is a pair `[encoded_audio, label]`, where `encoded_audio` is
a binary string whose encoding is specified in the summary metadata, and
`label` is a UTF-8 encoded Markdown string describing the audio clip.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

from tensorboard.compat import tf2 as tf
from tensorboard.plugins.audio import metadata
from tensorboard.util import lazy_tensor_creator


def audio(name,
          data,
          sample_rate,
          step=None,
          max_outputs=3,
          encoding=None,
          description=None):
  """Write an audio summary.

  Arguments:
    name: A name for this summary. The summary tag used for TensorBoard will
      be this name prefixed by any active name scopes.
    data: A `Tensor` representing audio data with shape `[k, t, c]`,
      where `k` is the number of audio clips, `t` is the number of
      frames, and `c` is the number of channels. Elements should be
      floating-point values in `[-1.0, 1.0]`. Any of the dimensions may
      be statically unknown (i.e., `None`).
    sample_rate: An `int` or rank-0 `int32` `Tensor` that represents the
      sample rate, in Hz. Must be positive.
    step: Explicit `int64`-castable monotonic step value for this summary. If
      omitted, this defaults to `tf.summary.experimental.get_step()`, which must
      not be None.
    max_outputs: Optional `int` or rank-0 integer `Tensor`. At most this
      many audio clips will be emitted at each step. When more than
      `max_outputs` many clips are provided, the first `max_outputs`
      many clips will be used and the rest silently discarded.
    encoding: Optional constant `str` for the desired encoding. Only "wav"
      is currently supported, but this is not guaranteed to remain the
      default, so if you want "wav" in particular, set this explicitly.
    description: Optional long-form description for this summary, as a
      constant `str`. Markdown is supported. Defaults to empty.

  Returns:
    True on success, or false if no summary was emitted because no default
    summary writer was available.

  Raises:
    ValueError: if a default writer exists, but no step was provided and
      `tf.summary.experimental.get_step()` is None.
  """
  audio_ops = getattr(tf, 'audio', None)
  if audio_ops is None:
    # Fallback for older versions of TF without tf.audio.
    from tensorflow.python.ops import gen_audio_ops as audio_ops

  if encoding is None:
    encoding = 'wav'
  if encoding != 'wav':
    raise ValueError('Unknown encoding: %r' % encoding)
  summary_metadata = metadata.create_summary_metadata(
      display_name=None,
      description=description,
      encoding=metadata.Encoding.Value('WAV'))
  inputs = [data, sample_rate, max_outputs, step]
  # TODO(https://github.com/tensorflow/tensorboard/issues/2109): remove fallback
  summary_scope = (
      getattr(tf.summary.experimental, 'summary_scope', None) or
      tf.summary.summary_scope)
  with summary_scope(
      name, 'audio_summary', values=inputs) as (tag, _):
    # Defer audio encoding preprocessing by passing it as a callable to write(),
    # wrapped in a LazyTensorCreator for backwards compatibility, so that we
    # only do this work when summaries are actually written.
    @lazy_tensor_creator.LazyTensorCreator
    def lazy_tensor():
      tf.debugging.assert_rank(data, 3)
      tf.debugging.assert_non_negative(max_outputs)
      limited_audio = data[:max_outputs]
      encode_fn = functools.partial(audio_ops.encode_wav,
                                    sample_rate=sample_rate)
      encoded_audio = tf.map_fn(encode_fn, limited_audio,
                                dtype=tf.string,
                                name='encode_each_audio')
      # Workaround for map_fn returning float dtype for an empty elems input.
      encoded_audio = tf.cond(
          tf.shape(input=encoded_audio)[0] > 0,
          lambda: encoded_audio, lambda: tf.constant([], tf.string))
      limited_labels = tf.tile([''], tf.shape(input=limited_audio)[:1])
      return tf.transpose(a=tf.stack([encoded_audio, limited_labels]))

    # To ensure that audio encoding logic is only executed when summaries
    # are written, we pass callable to `tensor` parameter.
    return tf.summary.write(
        tag=tag, tensor=lazy_tensor, step=step, metadata=summary_metadata)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper around plugin to conditionally enable it."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.plugins import base_plugin


class InteractiveInferencePluginLoader(base_plugin.TBLoader):
  """InteractiveInferencePlugin factory.

  This class checks for `tensorflow` install and dependency.
  """

  def load(self, context):
    """Returns the plugin, if possible.

    Args:
      context: The TBContext flags.

    Returns:
      A InteractiveInferencePlugin instance or None if it couldn't be loaded.
    """
    try:
      # pylint: disable=g-import-not-at-top,unused-import
      import tensorflow
    except ImportError:
      return
    # pylint: disable=line-too-long,g-import-not-at-top
    from tensorboard.plugins.interactive_inference.interactive_inference_plugin import InteractiveInferencePlugin
    return InteractiveInferencePlugin(context)



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Text summaries and TensorFlow operations to create them, V2 versions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorboard.compat import tf2 as tf
from tensorboard.compat.proto import summary_pb2
from tensorboard.plugins.text import metadata
from tensorboard.util import tensor_util


def text(name, data, step=None, description=None):
  """Write a text summary.

  Arguments:
    name: A name for this summary. The summary tag used for TensorBoard will
      be this name prefixed by any active name scopes.
    data: A UTF-8 string tensor value.
    step: Explicit `int64`-castable monotonic step value for this summary. If
      omitted, this defaults to `tf.summary.experimental.get_step()`, which must
      not be None.
    description: Optional long-form description for this summary, as a
      constant `str`. Markdown is supported. Defaults to empty.

  Returns:
    True on success, or false if no summary was emitted because no default
    summary writer was available.

  Raises:
    ValueError: if a default writer exists, but no step was provided and
      `tf.summary.experimental.get_step()` is None.
  """
  summary_metadata = metadata.create_summary_metadata(
      display_name=None, description=description)
  # TODO(https://github.com/tensorflow/tensorboard/issues/2109): remove fallback
  summary_scope = (
      getattr(tf.summary.experimental, 'summary_scope', None) or
      tf.summary.summary_scope)
  with summary_scope(
      name, 'text_summary', values=[data, step]) as (tag, _):
    tf.debugging.assert_type(data, tf.string)
    return tf.summary.write(
        tag=tag, tensor=data, step=step, metadata=summary_metadata)


def text_pb(tag, data, description=None):
  """Create a text tf.Summary protobuf.

  Arguments:
    tag: String tag for the summary.
    data: A Python bytestring (of type bytes), a Unicode string, or a numpy data
      array of those types.
    description: Optional long-form description for this summary, as a `str`.
      Markdown is supported. Defaults to empty.

  Raises:
    TypeError: If the type of the data is unsupported.

  Returns:
    A `tf.Summary` protobuf object.
  """
  try:
    tensor = tensor_util.make_tensor_proto(data, dtype=np.object)
  except TypeError as e:
    raise TypeError('tensor must be of type string', e)
  summary_metadata = metadata.create_summary_metadata(
      display_name=None, description=description)
  summary = summary_pb2.Summary()
  summary.value.add(tag=tag,
                    metadata=summary_metadata,
                    tensor=tensor)
  return summary



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# Lint as: python3
"""Provides authentication support for TensorBoardUploader."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import errno
import json
import os
import sys
import webbrowser

import google_auth_oauthlib.flow
import grpc
import google.auth
import google.auth.transport.requests
import google.oauth2.credentials

from tensorboard.uploader import util
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()


# OAuth2 scopes used for OpenID Connect:
# https://developers.google.com/identity/protocols/OpenIDConnect#scope-param
OPENID_CONNECT_SCOPES = (
    "openid",
    "https://www.googleapis.com/auth/userinfo.email",
)


# The client "secret" is public by design for installed apps. See
# https://developers.google.com/identity/protocols/OAuth2?csw=1#installed
OAUTH_CLIENT_CONFIG = u"""
{
  "installed": {
    "client_id": "373649185512-8v619h5kft38l4456nm2dj4ubeqsrvh6.apps.googleusercontent.com",
    "project_id": "hosted-tensorboard-prod",
    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
    "token_uri": "https://oauth2.googleapis.com/token",
    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
    "client_secret": "pOyAuU2yq2arsM98Bw5hwYtr",
    "redirect_uris": [
      "urn:ietf:wg:oauth:2.0:oob",
      "http://localhost"
    ]
  }
}
"""


# Components of the relative path (within the user settings directory) at which
# to store TensorBoard uploader credentials.
TENSORBOARD_CREDENTIALS_FILEPATH_PARTS = [
    "tensorboard", "credentials", "uploader-creds.json"]


class CredentialsStore(object):
  """Private file store for a `google.oauth2.credentials.Credentials`."""

  _DEFAULT_CONFIG_DIRECTORY = object()  # Sentinel value.

  def __init__(self, user_config_directory=_DEFAULT_CONFIG_DIRECTORY):
    """Creates a CredentialsStore.

    Args:
      user_config_directory: Optional absolute path to the root directory for
        storing user configs, under which to store the credentials file. If not
        set, defaults to a platform-specific location. If set to None, the
        store is disabled (reads return None; write and clear are no-ops).
    """
    if user_config_directory is CredentialsStore._DEFAULT_CONFIG_DIRECTORY:
      user_config_directory = util.get_user_config_directory()
      if user_config_directory is None:
        logger.warning(
            "Credentials caching disabled - no private config directory found")
    if user_config_directory is None:
      self._credentials_filepath = None
    else:
      self._credentials_filepath = os.path.join(
          user_config_directory, *TENSORBOARD_CREDENTIALS_FILEPATH_PARTS)

  def read_credentials(self):
    """Returns the current `google.oauth2.credentials.Credentials`, or None."""
    if self._credentials_filepath is None:
      return None
    if os.path.exists(self._credentials_filepath):
      return google.oauth2.credentials.Credentials.from_authorized_user_file(
          self._credentials_filepath)
    return None

  def write_credentials(self, credentials):
    """Writes a `google.oauth2.credentials.Credentials` to the store."""
    if not isinstance(credentials, google.oauth2.credentials.Credentials):
      raise TypeError("Cannot write credentials of type %s" % type(credentials))
    if self._credentials_filepath is None:
      return
    # Make the credential file private if not on Windows; on Windows we rely on
    # the default user config settings directory being private since we don't
    # have a straightforward way to make an individual file private.
    private = os.name != "nt"
    util.make_file_with_directories(self._credentials_filepath, private=private)
    data = {
        "refresh_token": credentials.refresh_token,
        "token_uri": credentials.token_uri,
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "scopes": credentials.scopes,
        "type": "authorized_user",
    }
    with open(self._credentials_filepath, "w") as f:
      json.dump(data, f)

  def clear(self):
    """Clears the store of any persisted credentials information."""
    if self._credentials_filepath is None:
      return
    try:
      os.remove(self._credentials_filepath)
    except OSError as e:
      if e.errno != errno.ENOENT:
        raise


def build_installed_app_flow(client_config):
  """Returns a `CustomInstalledAppFlow` for the given config.

  Args:
    client_config (Mapping[str, Any]): The client configuration in the Google
        client secrets format.

  Returns:
    CustomInstalledAppFlow: the constructed flow.
  """
  return CustomInstalledAppFlow.from_client_config(
      client_config, scopes=OPENID_CONNECT_SCOPES)


class CustomInstalledAppFlow(google_auth_oauthlib.flow.InstalledAppFlow):
  """Customized version of the Installed App OAuth2 flow."""

  def run(self, force_console=False):
    """Run the flow using a local server if possible, otherwise the console."""
    # TODO(b/141721828): make auto-detection smarter, especially for macOS.
    if not force_console and os.getenv("DISPLAY"):
      try:
        return self.run_local_server(port=0)
      except webbrowser.Error:
        sys.stderr.write("Falling back to console authentication flow...\n")
    return self.run_console()


class IdTokenAuthMetadataPlugin(grpc.AuthMetadataPlugin):
  """A `gRPC AuthMetadataPlugin` that uses ID tokens.

  This works like the existing `google.auth.transport.grpc.AuthMetadataPlugin`
  except that instead of always using access tokens, it preferentially uses the
  `Credentials.id_token` property if available (and logs an error otherwise).

  See http://www.grpc.io/grpc/python/grpc.html#grpc.AuthMetadataPlugin
  """

  def __init__(self, credentials, request):
    """Constructs an IdTokenAuthMetadataPlugin.

    Args:
      credentials (google.auth.credentials.Credentials): The credentials to
        add to requests.
      request (google.auth.transport.Request): A HTTP transport request object
        used to refresh credentials as needed.
    """
    super(IdTokenAuthMetadataPlugin, self).__init__()
    if not isinstance(credentials, google.oauth2.credentials.Credentials):
      raise TypeError(
          "Cannot get ID tokens from credentials type %s" % type(credentials))
    self._credentials = credentials
    self._request = request

  def __call__(self, context, callback):
    """Passes authorization metadata into the given callback.

    Args:
      context (grpc.AuthMetadataContext): The RPC context.
      callback (grpc.AuthMetadataPluginCallback): The callback that will
        be invoked to pass in the authorization metadata.
    """
    headers = {}
    self._credentials.before_request(
        self._request, context.method_name, context.service_url, headers)
    id_token = getattr(self._credentials, "id_token", None)
    if id_token:
      self._credentials.apply(headers, token=id_token)
    else:
      logger.error("Failed to find ID token credentials")
    # Pass headers as key-value pairs to match CallCredentials metadata.
    callback(list(headers.items()), None)


def id_token_call_credentials(credentials):
  """Constructs `grpc.CallCredentials` using `google.auth.Credentials.id_token`.

  Args:
    credentials (google.auth.credentials.Credentials): The credentials to use.

  Returns:
    grpc.CallCredentials: The call credentials.
  """
  request = google.auth.transport.requests.Request()
  return grpc.metadata_call_credentials(
      IdTokenAuthMetadataPlugin(credentials, request))



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Main program for the TensorBoard.dev uploader."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import json
import os
import sys
import textwrap

from absl import app
from absl import logging
from absl.flags import argparse_flags
import grpc
import six

from tensorboard.uploader import dev_creds
from tensorboard.uploader.proto import export_service_pb2
from tensorboard.uploader.proto import export_service_pb2_grpc
from tensorboard.uploader.proto import write_service_pb2_grpc
from tensorboard.uploader import auth
from tensorboard.uploader import exporter as exporter_lib
from tensorboard.uploader import server_info as server_info_lib
from tensorboard.uploader import uploader as uploader_lib
from tensorboard.uploader import util
from tensorboard.uploader.proto import server_info_pb2
from tensorboard import program
from tensorboard.plugins import base_plugin


# Temporary integration point for absl compatibility; will go away once
# migrated to TensorBoard subcommand.
_FLAGS = None


_MESSAGE_TOS = u"""\
Your use of this service is subject to Google's Terms of Service
<https://policies.google.com/terms> and Privacy Policy
<https://policies.google.com/privacy>, and TensorBoard.dev's Terms of Service
<https://tensorboard.dev/policy/terms/>.

This notice will not be shown again while you are logged into the uploader.
To log out, run `tensorboard dev auth revoke`.
"""


_SUBCOMMAND_FLAG = '_uploader__subcommand'
_SUBCOMMAND_KEY_UPLOAD = 'UPLOAD'
_SUBCOMMAND_KEY_DELETE = 'DELETE'
_SUBCOMMAND_KEY_LIST = 'LIST'
_SUBCOMMAND_KEY_EXPORT = 'EXPORT'
_SUBCOMMAND_KEY_AUTH = 'AUTH'
_AUTH_SUBCOMMAND_FLAG = '_uploader__subcommand_auth'
_AUTH_SUBCOMMAND_KEY_REVOKE = 'REVOKE'

_DEFAULT_ORIGIN = "https://tensorboard.dev"


def _prompt_for_user_ack(intent):
  """Prompts for user consent, exiting the program if they decline."""
  body = intent.get_ack_message_body()
  header = '\n***** TensorBoard Uploader *****\n'
  user_ack_message = '\n'.join((header, body, _MESSAGE_TOS))
  sys.stderr.write(user_ack_message)
  sys.stderr.write('\n')
  response = six.moves.input('Continue? (yes/NO) ')
  if response.lower() not in ('y', 'yes'):
    sys.exit(0)
  sys.stderr.write('\n')


def _define_flags(parser):
  """Configures flags on the provided argument parser.

  Integration point for `tensorboard.program`'s subcommand system.

  Args:
    parser: An `argparse.ArgumentParser` to be mutated.
  """

  subparsers = parser.add_subparsers()

  parser.add_argument(
      '--origin',
      type=str,
      default='',
      help='Experimental. Origin for TensorBoard.dev service to which '
      'to connect. If not set, defaults to %r.' % _DEFAULT_ORIGIN)

  parser.add_argument(
      '--api_endpoint',
      type=str,
      default='',
      help='Experimental. Direct URL for the API server accepting '
      'write requests. If set, will skip initial server handshake '
      'unless `--origin` is also set.')

  parser.add_argument(
      '--grpc_creds_type',
      type=str,
      default='ssl',
      choices=('local', 'ssl', 'ssl_dev'),
      help='The type of credentials to use for the gRPC client')

  parser.add_argument(
      '--auth_force_console',
      action='store_true',
      help='Set to true to force authentication flow to use the '
      '--console rather than a browser redirect to localhost.')

  upload = subparsers.add_parser(
      'upload', help='upload an experiment to TensorBoard.dev')
  upload.set_defaults(**{_SUBCOMMAND_FLAG: _SUBCOMMAND_KEY_UPLOAD})
  upload.add_argument(
      '--logdir',
      metavar='PATH',
      type=str,
      default=None,
      help='Directory containing the logs to process')

  delete = subparsers.add_parser(
      'delete',
      help='permanently delete an experiment',
      inherited_absl_flags=None)
  delete.set_defaults(**{_SUBCOMMAND_FLAG: _SUBCOMMAND_KEY_DELETE})
  # We would really like to call this next flag `--experiment` rather
  # than `--experiment_id`, but this is broken inside Google due to a
  # long-standing Python bug: <https://bugs.python.org/issue14365>
  # (Some Google-internal dependencies define `--experimental_*` flags.)
  # This isn't exactly a principled fix, but it gets the job done.
  delete.add_argument(
      '--experiment_id',
      metavar='EXPERIMENT_ID',
      type=str,
      default=None,
      help='ID of an experiment to delete permanently')

  list_parser = subparsers.add_parser(
      'list', help='list previously uploaded experiments')
  list_parser.set_defaults(**{_SUBCOMMAND_FLAG: _SUBCOMMAND_KEY_LIST})

  export = subparsers.add_parser(
      'export', help='download all your experiment data')
  export.set_defaults(**{_SUBCOMMAND_FLAG: _SUBCOMMAND_KEY_EXPORT})
  export.add_argument(
      '--outdir',
      metavar='OUTPUT_PATH',
      type=str,
      default=None,
      help='Directory into which to download all experiment data; '
      'must not yet exist')

  auth_parser = subparsers.add_parser('auth', help='log in, log out')
  auth_parser.set_defaults(**{_SUBCOMMAND_FLAG: _SUBCOMMAND_KEY_AUTH})
  auth_subparsers = auth_parser.add_subparsers()

  auth_revoke = auth_subparsers.add_parser(
      'revoke', help='revoke all existing credentials and log out')
  auth_revoke.set_defaults(
      **{_AUTH_SUBCOMMAND_FLAG: _AUTH_SUBCOMMAND_KEY_REVOKE})


def _parse_flags(argv=('',)):
  """Integration point for `absl.app`.

  Exits if flag values are invalid.

  Args:
    argv: CLI arguments, as with `sys.argv`, where the first argument is taken
      to be the name of the program being executed.

  Returns:
    Either argv[:1] if argv was non-empty, or [''] otherwise, as a mechanism
    for absl.app.run() compatibility.
  """
  parser = argparse_flags.ArgumentParser(
      prog='uploader',
      description=('Upload your TensorBoard experiments to TensorBoard.dev'))
  _define_flags(parser)
  arg0 = argv[0] if argv else ''
  global _FLAGS
  _FLAGS = parser.parse_args(argv[1:])
  return [arg0]


def _run(flags):
  """Runs the main uploader program given parsed flags.

  Args:
    flags: An `argparse.Namespace`.
  """

  logging.set_stderrthreshold(logging.WARNING)
  intent = _get_intent(flags)

  store = auth.CredentialsStore()
  if isinstance(intent, _AuthRevokeIntent):
    store.clear()
    sys.stderr.write('Logged out of uploader.\n')
    sys.stderr.flush()
    return
  # TODO(b/141723268): maybe reconfirm Google Account prior to reuse.
  credentials = store.read_credentials()
  if not credentials:
    _prompt_for_user_ack(intent)
    client_config = json.loads(auth.OAUTH_CLIENT_CONFIG)
    flow = auth.build_installed_app_flow(client_config)
    credentials = flow.run(force_console=flags.auth_force_console)
    sys.stderr.write('\n')  # Extra newline after auth flow messages.
    store.write_credentials(credentials)

  channel_options = None
  if flags.grpc_creds_type == 'local':
    channel_creds = grpc.local_channel_credentials()
  elif flags.grpc_creds_type == 'ssl':
    channel_creds = grpc.ssl_channel_credentials()
  elif flags.grpc_creds_type == 'ssl_dev':
    channel_creds = grpc.ssl_channel_credentials(dev_creds.DEV_SSL_CERT)
    channel_options = [('grpc.ssl_target_name_override', 'localhost')]
  else:
    msg = 'Invalid --grpc_creds_type %s' % flags.grpc_creds_type
    raise base_plugin.FlagsError(msg)

  try:
    server_info = _get_server_info(flags)
  except server_info_lib.CommunicationError as e:
    _die(str(e))
  _handle_server_info(server_info)

  if not server_info.api_server.endpoint:
    logging.error('Server info response: %s', server_info)
    _die('Internal error: frontend did not specify an API server')
  composite_channel_creds = grpc.composite_channel_credentials(
      channel_creds, auth.id_token_call_credentials(credentials))

  # TODO(@nfelt): In the `_UploadIntent` case, consider waiting until
  # logdir exists to open channel.
  channel = grpc.secure_channel(
      server_info.api_server.endpoint,
      composite_channel_creds,
      options=channel_options)
  with channel:
    intent.execute(server_info, channel)


@six.add_metaclass(abc.ABCMeta)
class _Intent(object):
  """A description of the user's intent in invoking this program.

  Each valid set of CLI flags corresponds to one intent: e.g., "upload
  data from this logdir", or "delete the experiment with that ID".
  """

  @abc.abstractmethod
  def get_ack_message_body(self):
    """Gets the message to show when executing this intent at first login.

    This need not include the header (program name) or Terms of Service
    notice.

    Returns:
      A Unicode string, potentially spanning multiple lines.
    """
    pass

  @abc.abstractmethod
  def execute(self, server_info, channel):
    """Carries out this intent with the specified gRPC channel.

    Args:
      server_info: A `server_info_pb2.ServerInfoResponse` value.
      channel: A connected gRPC channel whose server provides the TensorBoard
        reader and writer services.
    """
    pass


class _AuthRevokeIntent(_Intent):
  """The user intends to revoke credentials."""

  def get_ack_message_body(self):
    """Must not be called."""
    raise AssertionError('No user ack needed to revoke credentials')

  def execute(self, server_info, channel):
    """Execute handled specially by `main`. Must not be called."""
    raise AssertionError('_AuthRevokeIntent should not be directly executed')


class _DeleteExperimentIntent(_Intent):
  """The user intends to delete an experiment."""

  _MESSAGE_TEMPLATE = textwrap.dedent(u"""\
      This will delete the experiment on https://tensorboard.dev with the
      following experiment ID:

      {experiment_id}

      You have chosen to delete an experiment. All experiments uploaded
      to TensorBoard.dev are publicly visible. Do not upload sensitive
      data.
  """)

  def __init__(self, experiment_id):
    self.experiment_id = experiment_id

  def get_ack_message_body(self):
    return self._MESSAGE_TEMPLATE.format(experiment_id=self.experiment_id)

  def execute(self, server_info, channel):
    api_client = write_service_pb2_grpc.TensorBoardWriterServiceStub(channel)
    experiment_id = self.experiment_id
    if not experiment_id:
      raise base_plugin.FlagsError(
          'Must specify a non-empty experiment ID to delete.')
    try:
      uploader_lib.delete_experiment(api_client, experiment_id)
    except uploader_lib.ExperimentNotFoundError:
      _die(
          'No such experiment %s. Either it never existed or it has '
          'already been deleted.' % experiment_id)
    except uploader_lib.PermissionDeniedError:
      _die(
          'Cannot delete experiment %s because it is owned by a '
          'different user.' % experiment_id)
    except grpc.RpcError as e:
      _die('Internal error deleting experiment: %s' % e)
    print('Deleted experiment %s.' % experiment_id)


class _ListIntent(_Intent):
  """The user intends to list all their experiments."""

  _MESSAGE = textwrap.dedent(u"""\
      This will list all experiments that you've uploaded to
      https://tensorboard.dev. TensorBoard.dev experiments are visible
      to everyone. Do not upload sensitive data.
  """)

  def get_ack_message_body(self):
    return self._MESSAGE

  def execute(self, server_info, channel):
    api_client = export_service_pb2_grpc.TensorBoardExporterServiceStub(channel)
    fieldmask = export_service_pb2.ExperimentMask(
        create_time=True,
        update_time=True,
        num_scalars=True,
        num_runs=True,
        num_tags=True,
    )
    gen = exporter_lib.list_experiments(api_client, fieldmask=fieldmask)
    count = 0
    for experiment in gen:
      count += 1
      if not isinstance(experiment, export_service_pb2.Experiment):
        url = server_info_lib.experiment_url(server_info, experiment)
        print(url)
        continue
      experiment_id = experiment.experiment_id
      url = server_info_lib.experiment_url(server_info, experiment_id)
      print(url)
      data = [
          ('Id', experiment.experiment_id),
          ('Created', util.format_time(experiment.create_time)),
          ('Updated', util.format_time(experiment.update_time)),
          ('Scalars', str(experiment.num_scalars)),
          ('Runs', str(experiment.num_runs)),
          ('Tags', str(experiment.num_tags)),
      ]
      for (name, value) in data:
        print('\t%s %s' % (name.ljust(10), value))
    sys.stdout.flush()
    if not count:
      sys.stderr.write(
          'No experiments. Use `tensorboard dev upload` to get started.\n')
    else:
      sys.stderr.write('Total: %d experiment(s)\n' % count)
    sys.stderr.flush()


class _UploadIntent(_Intent):
  """The user intends to upload an experiment from the given logdir."""

  _MESSAGE_TEMPLATE = textwrap.dedent(u"""\
      This will upload your TensorBoard logs to https://tensorboard.dev/ from
      the following directory:

      {logdir}

      This TensorBoard will be visible to everyone. Do not upload sensitive
      data.
  """)

  def __init__(self, logdir):
    self.logdir = logdir

  def get_ack_message_body(self):
    return self._MESSAGE_TEMPLATE.format(logdir=self.logdir)

  def execute(self, server_info, channel):
    api_client = write_service_pb2_grpc.TensorBoardWriterServiceStub(channel)
    uploader = uploader_lib.TensorBoardUploader(api_client, self.logdir)
    experiment_id = uploader.create_experiment()
    url = server_info_lib.experiment_url(server_info, experiment_id)
    print("Upload started and will continue reading any new data as it's added")
    print("to the logdir. To stop uploading, press Ctrl-C.")
    print("View your TensorBoard live at: %s" % url)
    try:
      uploader.start_uploading()
    except uploader_lib.ExperimentNotFoundError:
      print('Experiment was deleted; uploading has been cancelled')
      return
    except KeyboardInterrupt:
      print()
      print('Upload stopped. View your TensorBoard at %s' % url)
      return
    # TODO(@nfelt): make it possible for the upload cycle to end once we
    #   detect that no more runs are active, so this code can be reached.
    print('Done! View your TensorBoard at %s' % url)


class _ExportIntent(_Intent):
  """The user intends to download all their experiment data."""

  _MESSAGE_TEMPLATE = textwrap.dedent(u"""\
      This will download all your experiment data from https://tensorboard.dev
      and save it to the following directory:

      {output_dir}

      Downloading your experiment data does not delete it from the
      service. All experiments uploaded to TensorBoard.dev are publicly
      visible. Do not upload sensitive data.
  """)

  def __init__(self, output_dir):
    self.output_dir = output_dir

  def get_ack_message_body(self):
    return self._MESSAGE_TEMPLATE.format(output_dir=self.output_dir)

  def execute(self, server_info, channel):
    api_client = export_service_pb2_grpc.TensorBoardExporterServiceStub(channel)
    outdir = self.output_dir
    try:
      exporter = exporter_lib.TensorBoardExporter(api_client, outdir)
    except exporter_lib.OutputDirectoryExistsError:
      msg = 'Output directory already exists: %r' % outdir
      raise base_plugin.FlagsError(msg)
    num_experiments = 0
    try:
      for experiment_id in exporter.export():
        num_experiments += 1
        print('Downloaded experiment %s' % experiment_id)
    except exporter_lib.GrpcTimeoutException as e:
      print(
        '\nUploader has failed because of a timeout error.  Please reach '
        'out via e-mail to tensorboard.dev-support@google.com to get help '
        'completing your export of experiment %s.' % e.experiment_id)
    print('Done. Downloaded %d experiments to: %s' % (num_experiments, outdir))


def _get_intent(flags):
  """Determines what the program should do (upload, delete, ...).

  Args:
    flags: An `argparse.Namespace` with the parsed flags.

  Returns:
    An `_Intent` instance.

  Raises:
    base_plugin.FlagsError: If the command-line `flags` do not correctly
      specify an intent.
  """
  cmd = getattr(flags, _SUBCOMMAND_FLAG, None)
  if cmd is None:
    raise base_plugin.FlagsError('Must specify subcommand (try --help).')
  if cmd == _SUBCOMMAND_KEY_UPLOAD:
    if flags.logdir:
      return _UploadIntent(os.path.expanduser(flags.logdir))
    else:
      raise base_plugin.FlagsError(
          'Must specify directory to upload via `--logdir`.')
  elif cmd == _SUBCOMMAND_KEY_DELETE:
    if flags.experiment_id:
      return _DeleteExperimentIntent(flags.experiment_id)
    else:
      raise base_plugin.FlagsError(
          'Must specify experiment to delete via `--experiment_id`.')
  elif cmd == _SUBCOMMAND_KEY_LIST:
    return _ListIntent()
  elif cmd == _SUBCOMMAND_KEY_EXPORT:
    if flags.outdir:
      return _ExportIntent(flags.outdir)
    else:
      raise base_plugin.FlagsError(
          'Must specify output directory via `--outdir`.')
  elif cmd == _SUBCOMMAND_KEY_AUTH:
    auth_cmd = getattr(flags, _AUTH_SUBCOMMAND_FLAG, None)
    if auth_cmd is None:
      raise base_plugin.FlagsError('Must specify a subcommand to `auth`.')
    if auth_cmd == _AUTH_SUBCOMMAND_KEY_REVOKE:
      return _AuthRevokeIntent()
    else:
      raise AssertionError('Unknown auth subcommand %r' % (auth_cmd,))
  else:
    raise AssertionError('Unknown subcommand %r' % (cmd,))


def _get_server_info(flags):
  origin = flags.origin or _DEFAULT_ORIGIN
  if flags.api_endpoint and not flags.origin:
    return server_info_lib.create_server_info(origin, flags.api_endpoint)
  server_info = server_info_lib.fetch_server_info(origin)
  # Override with any API server explicitly specified on the command
  # line, but only if the server accepted our initial handshake.
  if flags.api_endpoint and server_info.api_server.endpoint:
    server_info.api_server.endpoint = flags.api_endpoint
  return server_info


def _handle_server_info(info):
  compat = info.compatibility
  if compat.verdict == server_info_pb2.VERDICT_WARN:
    sys.stderr.write('Warning [from server]: %s\n' % compat.details)
    sys.stderr.flush()
  elif compat.verdict == server_info_pb2.VERDICT_ERROR:
    _die('Error [from server]: %s' % compat.details)
  else:
    # OK or unknown; assume OK.
    if compat.details:
      sys.stderr.write('%s\n' % compat.details)
      sys.stderr.flush()


def _die(message):
  sys.stderr.write('%s\n' % (message,))
  sys.stderr.flush()
  sys.exit(1)


def main(unused_argv):
  global _FLAGS
  flags = _FLAGS
  # Prevent accidental use of `_FLAGS` until migration to TensorBoard
  # subcommand is complete, at which point `_FLAGS` goes away.
  del _FLAGS
  return _run(flags)


class UploaderSubcommand(program.TensorBoardSubcommand):
  """Integration point with `tensorboard` CLI."""

  def name(self):
    return 'dev'

  def define_flags(self, parser):
    _define_flags(parser)

  def run(self, flags):
    return _run(flags)

  def help(self):
    return 'upload data to TensorBoard.dev'


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorboard/uploader/proto/write_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorboard.uploader.proto import scalar_pb2 as tensorboard_dot_uploader_dot_proto_dot_scalar__pb2
from tensorboard.compat.proto import summary_pb2 as tensorboard_dot_compat_dot_proto_dot_summary__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorboard/uploader/proto/write_service.proto',
  package='tensorboard.service',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n.tensorboard/uploader/proto/write_service.proto\x12\x13tensorboard.service\x1a\'tensorboard/uploader/proto/scalar.proto\x1a&tensorboard/compat/proto/summary.proto\"\x19\n\x17\x43reateExperimentRequest\">\n\x18\x43reateExperimentResponse\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\"0\n\x17\x44\x65leteExperimentRequest\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\"\x1a\n\x18\x44\x65leteExperimentResponse\"\'\n\x10PurgeDataRequest\x12\x13\n\x0b\x62\x61tch_limit\x18\x01 \x01(\x05\"I\n\x11PurgeDataResponse\x12\x34\n\x0bpurge_stats\x18\x01 \x01(\x0b\x32\x1f.tensorboard.service.PurgeStats\">\n\nPurgeStats\x12\x0c\n\x04tags\x18\x01 \x01(\x05\x12\x13\n\x0b\x65xperiments\x18\x02 \x01(\x05\x12\r\n\x05users\x18\x03 \x01(\x05\"\xad\x02\n\x12WriteScalarRequest\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x39\n\x04runs\x18\x02 \x03(\x0b\x32+.tensorboard.service.WriteScalarRequest.Run\x1aN\n\x03Run\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x39\n\x04tags\x18\x02 \x03(\x0b\x32+.tensorboard.service.WriteScalarRequest.Tag\x1au\n\x03Tag\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x30\n\x06points\x18\x02 \x03(\x0b\x32 .tensorboard.service.ScalarPoint\x12.\n\x08metadata\x18\x03 \x01(\x0b\x32\x1c.tensorboard.SummaryMetadata\"\x15\n\x13WriteScalarResponse\"\x16\n\x14\x44\x65leteOwnUserRequest\"\x17\n\x15\x44\x65leteOwnUserResponse2\xac\x04\n\x18TensorBoardWriterService\x12q\n\x10\x43reateExperiment\x12,.tensorboard.service.CreateExperimentRequest\x1a-.tensorboard.service.CreateExperimentResponse\"\x00\x12q\n\x10\x44\x65leteExperiment\x12,.tensorboard.service.DeleteExperimentRequest\x1a-.tensorboard.service.DeleteExperimentResponse\"\x00\x12\\\n\tPurgeData\x12%.tensorboard.service.PurgeDataRequest\x1a&.tensorboard.service.PurgeDataResponse\"\x00\x12\x62\n\x0bWriteScalar\x12\'.tensorboard.service.WriteScalarRequest\x1a(.tensorboard.service.WriteScalarResponse\"\x00\x12h\n\rDeleteOwnUser\x12).tensorboard.service.DeleteOwnUserRequest\x1a*.tensorboard.service.DeleteOwnUserResponse\"\x00\x62\x06proto3')
  ,
  dependencies=[tensorboard_dot_uploader_dot_proto_dot_scalar__pb2.DESCRIPTOR,tensorboard_dot_compat_dot_proto_dot_summary__pb2.DESCRIPTOR,])




_CREATEEXPERIMENTREQUEST = _descriptor.Descriptor(
  name='CreateExperimentRequest',
  full_name='tensorboard.service.CreateExperimentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=177,
)


_CREATEEXPERIMENTRESPONSE = _descriptor.Descriptor(
  name='CreateExperimentResponse',
  full_name='tensorboard.service.CreateExperimentResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='tensorboard.service.CreateExperimentResponse.experiment_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='url', full_name='tensorboard.service.CreateExperimentResponse.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=179,
  serialized_end=241,
)


_DELETEEXPERIMENTREQUEST = _descriptor.Descriptor(
  name='DeleteExperimentRequest',
  full_name='tensorboard.service.DeleteExperimentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='tensorboard.service.DeleteExperimentRequest.experiment_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=243,
  serialized_end=291,
)


_DELETEEXPERIMENTRESPONSE = _descriptor.Descriptor(
  name='DeleteExperimentResponse',
  full_name='tensorboard.service.DeleteExperimentResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=293,
  serialized_end=319,
)


_PURGEDATAREQUEST = _descriptor.Descriptor(
  name='PurgeDataRequest',
  full_name='tensorboard.service.PurgeDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_limit', full_name='tensorboard.service.PurgeDataRequest.batch_limit', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=360,
)


_PURGEDATARESPONSE = _descriptor.Descriptor(
  name='PurgeDataResponse',
  full_name='tensorboard.service.PurgeDataResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='purge_stats', full_name='tensorboard.service.PurgeDataResponse.purge_stats', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=362,
  serialized_end=435,
)


_PURGESTATS = _descriptor.Descriptor(
  name='PurgeStats',
  full_name='tensorboard.service.PurgeStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tags', full_name='tensorboard.service.PurgeStats.tags', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='experiments', full_name='tensorboard.service.PurgeStats.experiments', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='users', full_name='tensorboard.service.PurgeStats.users', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=437,
  serialized_end=499,
)


_WRITESCALARREQUEST_RUN = _descriptor.Descriptor(
  name='Run',
  full_name='tensorboard.service.WriteScalarRequest.Run',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tensorboard.service.WriteScalarRequest.Run.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tags', full_name='tensorboard.service.WriteScalarRequest.Run.tags', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=606,
  serialized_end=684,
)

_WRITESCALARREQUEST_TAG = _descriptor.Descriptor(
  name='Tag',
  full_name='tensorboard.service.WriteScalarRequest.Tag',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tensorboard.service.WriteScalarRequest.Tag.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='points', full_name='tensorboard.service.WriteScalarRequest.Tag.points', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='tensorboard.service.WriteScalarRequest.Tag.metadata', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=686,
  serialized_end=803,
)

_WRITESCALARREQUEST = _descriptor.Descriptor(
  name='WriteScalarRequest',
  full_name='tensorboard.service.WriteScalarRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='tensorboard.service.WriteScalarRequest.experiment_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='runs', full_name='tensorboard.service.WriteScalarRequest.runs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_WRITESCALARREQUEST_RUN, _WRITESCALARREQUEST_TAG, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=502,
  serialized_end=803,
)


_WRITESCALARRESPONSE = _descriptor.Descriptor(
  name='WriteScalarResponse',
  full_name='tensorboard.service.WriteScalarResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=805,
  serialized_end=826,
)


_DELETEOWNUSERREQUEST = _descriptor.Descriptor(
  name='DeleteOwnUserRequest',
  full_name='tensorboard.service.DeleteOwnUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=828,
  serialized_end=850,
)


_DELETEOWNUSERRESPONSE = _descriptor.Descriptor(
  name='DeleteOwnUserResponse',
  full_name='tensorboard.service.DeleteOwnUserResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=852,
  serialized_end=875,
)

_PURGEDATARESPONSE.fields_by_name['purge_stats'].message_type = _PURGESTATS
_WRITESCALARREQUEST_RUN.fields_by_name['tags'].message_type = _WRITESCALARREQUEST_TAG
_WRITESCALARREQUEST_RUN.containing_type = _WRITESCALARREQUEST
_WRITESCALARREQUEST_TAG.fields_by_name['points'].message_type = tensorboard_dot_uploader_dot_proto_dot_scalar__pb2._SCALARPOINT
_WRITESCALARREQUEST_TAG.fields_by_name['metadata'].message_type = tensorboard_dot_compat_dot_proto_dot_summary__pb2._SUMMARYMETADATA
_WRITESCALARREQUEST_TAG.containing_type = _WRITESCALARREQUEST
_WRITESCALARREQUEST.fields_by_name['runs'].message_type = _WRITESCALARREQUEST_RUN
DESCRIPTOR.message_types_by_name['CreateExperimentRequest'] = _CREATEEXPERIMENTREQUEST
DESCRIPTOR.message_types_by_name['CreateExperimentResponse'] = _CREATEEXPERIMENTRESPONSE
DESCRIPTOR.message_types_by_name['DeleteExperimentRequest'] = _DELETEEXPERIMENTREQUEST
DESCRIPTOR.message_types_by_name['DeleteExperimentResponse'] = _DELETEEXPERIMENTRESPONSE
DESCRIPTOR.message_types_by_name['PurgeDataRequest'] = _PURGEDATAREQUEST
DESCRIPTOR.message_types_by_name['PurgeDataResponse'] = _PURGEDATARESPONSE
DESCRIPTOR.message_types_by_name['PurgeStats'] = _PURGESTATS
DESCRIPTOR.message_types_by_name['WriteScalarRequest'] = _WRITESCALARREQUEST
DESCRIPTOR.message_types_by_name['WriteScalarResponse'] = _WRITESCALARRESPONSE
DESCRIPTOR.message_types_by_name['DeleteOwnUserRequest'] = _DELETEOWNUSERREQUEST
DESCRIPTOR.message_types_by_name['DeleteOwnUserResponse'] = _DELETEOWNUSERRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CreateExperimentRequest = _reflection.GeneratedProtocolMessageType('CreateExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEEXPERIMENTREQUEST,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.CreateExperimentRequest)
  })
_sym_db.RegisterMessage(CreateExperimentRequest)

CreateExperimentResponse = _reflection.GeneratedProtocolMessageType('CreateExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEEXPERIMENTRESPONSE,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.CreateExperimentResponse)
  })
_sym_db.RegisterMessage(CreateExperimentResponse)

DeleteExperimentRequest = _reflection.GeneratedProtocolMessageType('DeleteExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEEXPERIMENTREQUEST,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.DeleteExperimentRequest)
  })
_sym_db.RegisterMessage(DeleteExperimentRequest)

DeleteExperimentResponse = _reflection.GeneratedProtocolMessageType('DeleteExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETEEXPERIMENTRESPONSE,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.DeleteExperimentResponse)
  })
_sym_db.RegisterMessage(DeleteExperimentResponse)

PurgeDataRequest = _reflection.GeneratedProtocolMessageType('PurgeDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _PURGEDATAREQUEST,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.PurgeDataRequest)
  })
_sym_db.RegisterMessage(PurgeDataRequest)

PurgeDataResponse = _reflection.GeneratedProtocolMessageType('PurgeDataResponse', (_message.Message,), {
  'DESCRIPTOR' : _PURGEDATARESPONSE,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.PurgeDataResponse)
  })
_sym_db.RegisterMessage(PurgeDataResponse)

PurgeStats = _reflection.GeneratedProtocolMessageType('PurgeStats', (_message.Message,), {
  'DESCRIPTOR' : _PURGESTATS,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.PurgeStats)
  })
_sym_db.RegisterMessage(PurgeStats)

WriteScalarRequest = _reflection.GeneratedProtocolMessageType('WriteScalarRequest', (_message.Message,), {

  'Run' : _reflection.GeneratedProtocolMessageType('Run', (_message.Message,), {
    'DESCRIPTOR' : _WRITESCALARREQUEST_RUN,
    '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
    # @@protoc_insertion_point(class_scope:tensorboard.service.WriteScalarRequest.Run)
    })
  ,

  'Tag' : _reflection.GeneratedProtocolMessageType('Tag', (_message.Message,), {
    'DESCRIPTOR' : _WRITESCALARREQUEST_TAG,
    '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
    # @@protoc_insertion_point(class_scope:tensorboard.service.WriteScalarRequest.Tag)
    })
  ,
  'DESCRIPTOR' : _WRITESCALARREQUEST,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.WriteScalarRequest)
  })
_sym_db.RegisterMessage(WriteScalarRequest)
_sym_db.RegisterMessage(WriteScalarRequest.Run)
_sym_db.RegisterMessage(WriteScalarRequest.Tag)

WriteScalarResponse = _reflection.GeneratedProtocolMessageType('WriteScalarResponse', (_message.Message,), {
  'DESCRIPTOR' : _WRITESCALARRESPONSE,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.WriteScalarResponse)
  })
_sym_db.RegisterMessage(WriteScalarResponse)

DeleteOwnUserRequest = _reflection.GeneratedProtocolMessageType('DeleteOwnUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEOWNUSERREQUEST,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.DeleteOwnUserRequest)
  })
_sym_db.RegisterMessage(DeleteOwnUserRequest)

DeleteOwnUserResponse = _reflection.GeneratedProtocolMessageType('DeleteOwnUserResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETEOWNUSERRESPONSE,
  '__module__' : 'tensorboard.uploader.proto.write_service_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.service.DeleteOwnUserResponse)
  })
_sym_db.RegisterMessage(DeleteOwnUserResponse)



_TENSORBOARDWRITERSERVICE = _descriptor.ServiceDescriptor(
  name='TensorBoardWriterService',
  full_name='tensorboard.service.TensorBoardWriterService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=878,
  serialized_end=1434,
  methods=[
  _descriptor.MethodDescriptor(
    name='CreateExperiment',
    full_name='tensorboard.service.TensorBoardWriterService.CreateExperiment',
    index=0,
    containing_service=None,
    input_type=_CREATEEXPERIMENTREQUEST,
    output_type=_CREATEEXPERIMENTRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteExperiment',
    full_name='tensorboard.service.TensorBoardWriterService.DeleteExperiment',
    index=1,
    containing_service=None,
    input_type=_DELETEEXPERIMENTREQUEST,
    output_type=_DELETEEXPERIMENTRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PurgeData',
    full_name='tensorboard.service.TensorBoardWriterService.PurgeData',
    index=2,
    containing_service=None,
    input_type=_PURGEDATAREQUEST,
    output_type=_PURGEDATARESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='WriteScalar',
    full_name='tensorboard.service.TensorBoardWriterService.WriteScalar',
    index=3,
    containing_service=None,
    input_type=_WRITESCALARREQUEST,
    output_type=_WRITESCALARRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteOwnUser',
    full_name='tensorboard.service.TensorBoardWriterService.DeleteOwnUser',
    index=4,
    containing_service=None,
    input_type=_DELETEOWNUSERREQUEST,
    output_type=_DELETEOWNUSERRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_TENSORBOARDWRITERSERVICE)

DESCRIPTOR.services_by_name['TensorBoardWriterService'] = _TENSORBOARDWRITERSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Iterator adapter that supports peeking ahead."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class PeekableIterator(object):
  """Iterator adapter that supports peeking ahead.

  As with most Python iterators, this is also iterable; its `__iter__`
  returns itself.

  This class is not thread-safe. Use external synchronization if
  iterating concurrently.
  """

  def __init__(self, iterable):
    """Initializes a peeking iterator wrapping the provided iterable.

    Args:
      iterable: An iterable to wrap.
    """
    self._iterator = iter(iterable)
    self._has_peeked = False
    self._peeked_element = None

  def has_next(self):
    """Checks whether there are any more items in this iterator.

    The next call to `next` or `peek` will raise `StopIteration` if and
    only if this method returns `False`.

    Returns:
      `True` if there are any more items in this iterator, else `False`.
    """
    try:
      self.peek()
      return True
    except StopIteration:
      return False

  def peek(self):
    """Gets the next item in the iterator without consuming it.

    Multiple consecutive calls will return the same element.

    Returns:
      The value that would be returned by `next`.

    Raises:
      StopIteration: If there are no more items in the iterator.
    """
    if not self._has_peeked:
      self._peeked_element = next(self._iterator)
      self._has_peeked = True
    return self._peeked_element

  def __iter__(self):
    return self

  def __next__(self):
    if self._has_peeked:
      self._has_peeked = False
      result = self._peeked_element
      self._peeked_element = None  # allow GC
      return result
    else:
      return next(self._iterator)

  def next(self):
    # (Like `__next__`, but Python 2.)
    return self.__next__()



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Downloads experiment data from TensorBoard.dev."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import base64
import errno
import grpc
import json
import os
import string
import time

import six

from tensorboard.uploader.proto import export_service_pb2
from tensorboard.uploader import util
from tensorboard.util import grpc_util

# Characters that are assumed to be safe in filenames. Note that the
# server's experiment IDs are base64 encodings of 16-byte blobs, so they
# can theoretically collide on case-insensitive filesystems. Each
# character has a ~3% chance of colliding, and so two random IDs have
# about a ~10^-33 chance of colliding. As a precaution, we'll still
# detect collision and fail fast rather than overwriting data.
_FILENAME_SAFE_CHARS = frozenset(string.ascii_letters + string.digits + "-_")

# Maximum value of a signed 64-bit integer.
_MAX_INT64 = 2**63 - 1

class TensorBoardExporter(object):
  """Exports all of the user's experiment data from TensorBoard.dev.

  Data is exported into a directory, with one file per experiment. Each
  experiment file is a sequence of time series, represented as a stream
  of JSON objects, one per line. Each JSON object includes a run name,
  tag name, `tensorboard.compat.proto.summary_pb2.SummaryMetadata` proto
  (base64-encoded, standard RFC 4648 alphabet), and set of points.
  Points are stored in three equal-length lists of steps, wall times (as
  seconds since epoch), and scalar values, for storage efficiency.

  Such streams of JSON objects may be conveniently processed with tools
  like jq(1).

  For example one line of an experiment file might read (when
  pretty-printed):

      {
        "points": {
          "steps": [0, 5],
          "values": [4.8935227394104, 2.5438034534454346],
          "wall_times": [1563406522.669238, 1563406523.0268838]
        },
        "run": "lr_1E-04,conv=1,fc=2",
        "summary_metadata": "CgkKB3NjYWxhcnMSC3hlbnQveGVudF8x",
        "tag": "xent/xent_1"
      }

  This is a time series with two points, both logged on 2019-07-17, one
  about 0.36 seconds after the other.
  """

  def __init__(self, reader_service_client, output_directory):
    """Constructs a TensorBoardExporter.

    Args:
      reader_service_client: A TensorBoardExporterService stub instance.
      output_directory: Path to a directory into which to write data. The
        directory must not exist, to avoid stomping existing or concurrent
        output. Its ancestors will be created if needed.
    """
    self._api = reader_service_client
    self._outdir = output_directory
    parent_dir = os.path.dirname(self._outdir)
    if parent_dir:
      _mkdir_p(parent_dir)
    try:
      os.mkdir(self._outdir)
    except OSError as e:
      if e.errno == errno.EEXIST:
        # Bail to avoid stomping existing output.
        raise OutputDirectoryExistsError()

  def export(self, read_time=None):
    """Executes the export flow.

    Args:
      read_time: A fixed timestamp from which to export data, as float seconds
        since epoch (like `time.time()`). Optional; defaults to the current
        time.

    Yields:
      After each experiment is successfully downloaded, the ID of that
      experiment, as a string.
    """
    if read_time is None:
      read_time = time.time()
    for experiment_id in self._request_experiment_ids(read_time):
      filepath = _scalars_filepath(self._outdir, experiment_id)
      try:
        with _open_excl(filepath) as outfile:
          data = self._request_scalar_data(experiment_id, read_time)
          for block in data:
            json.dump(block, outfile, sort_keys=True)
            outfile.write("\n")
            outfile.flush()
        yield experiment_id
      except grpc.RpcError as e:
        if e.code() == grpc.StatusCode.CANCELLED:
          raise GrpcTimeoutException(experiment_id)
        else:
          raise

  def _request_experiment_ids(self, read_time):
    """Yields all of the calling user's experiment IDs, as strings."""
    for experiment in list_experiments(self._api, read_time=read_time):
      if isinstance(experiment, export_service_pb2.Experiment):
        yield experiment.experiment_id
      elif isinstance(experiment, six.string_types):
        yield experiment
      else:
        raise AssertionError("Unexpected experiment type: %r" % (experiment,))

  def _request_scalar_data(self, experiment_id, read_time):
    """Yields JSON-serializable blocks of scalar data."""
    request = export_service_pb2.StreamExperimentDataRequest()
    request.experiment_id = experiment_id
    util.set_timestamp(request.read_timestamp, read_time)
    # No special error handling as we don't expect any errors from these
    # calls: all experiments should exist (read consistency timestamp)
    # and be owned by the calling user (only queried for own experiment
    # IDs). Any non-transient errors would be internal, and we have no
    # way to efficiently resume from transient errors because the server
    # does not support pagination.
    stream = self._api.StreamExperimentData(
        request, metadata=grpc_util.version_metadata())
    for response in stream:
      metadata = base64.b64encode(
          response.tag_metadata.SerializeToString()).decode("ascii")
      wall_times = [t.ToNanoseconds() / 1e9 for t in response.points.wall_times]
      yield {
          u"run": response.run_name,
          u"tag": response.tag_name,
          u"summary_metadata": metadata,
          u"points": {
              u"steps": list(response.points.steps),
              u"wall_times": wall_times,
              u"values": list(response.points.values),
          },
      }


def list_experiments(api_client, fieldmask=None, read_time=None):
  """Yields all of the calling user's experiments.

  Args:
    api_client: A TensorBoardExporterService stub instance.
    fieldmask: An optional `export_service_pb2.ExperimentMask` value.
    read_time: A fixed timestamp from which to export data, as float seconds
      since epoch (like `time.time()`). Optional; defaults to the current
      time.

  Yields:
    For each experiment owned by the user, an `export_service_pb2.Experiment`
    value, or a simple string experiment ID for older servers.
  """
  if read_time is None:
    read_time = time.time()
  request = export_service_pb2.StreamExperimentsRequest(limit=_MAX_INT64)
  util.set_timestamp(request.read_timestamp, read_time)
  if fieldmask:
    request.experiments_mask.CopyFrom(fieldmask)
  stream = api_client.StreamExperiments(
      request, metadata=grpc_util.version_metadata())
  for response in stream:
    if response.experiments:
      for experiment in response.experiments:
        yield experiment
    else:
      # Old servers.
      for experiment_id in response.experiment_ids:
        yield experiment_id


class OutputDirectoryExistsError(ValueError):
  pass


class OutputFileExistsError(ValueError):
  # Like Python 3's `__builtins__.FileExistsError`.
  pass

class GrpcTimeoutException(Exception):
  def __init__(self, experiment_id):
    super(GrpcTimeoutException, self).__init__(experiment_id)
    self.experiment_id = experiment_id

def _scalars_filepath(base_dir, experiment_id):
  """Gets file path in which to store scalars for the given experiment."""
  # Experiment IDs from the server should be filename-safe; verify
  # this before creating any files.
  bad_chars = frozenset(experiment_id) - _FILENAME_SAFE_CHARS
  if bad_chars:
    raise RuntimeError(
        "Unexpected characters ({bad_chars!r}) in experiment ID {eid!r}".format(
            bad_chars=sorted(bad_chars), eid=experiment_id))
  return os.path.join(base_dir, "scalars_%s.json" % experiment_id)


def _mkdir_p(path):
  """Like `os.makedirs(path, exist_ok=True)`, but Python 2-compatible."""
  try:
    os.makedirs(path)
  except OSError as e:
    if e.errno != errno.EEXIST or not os.path.isdir(path):
      raise


def _open_excl(path):
  """Like `open(path, "x")`, but Python 2-compatible."""
  try:
    # `os.O_EXCL` works on Windows as well as POSIX-compliant systems.
    # See: <https://bugs.python.org/issue12760>
    fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
  except OSError as e:
    if e.errno == errno.EEXIST:
      raise OutputFileExistsError(path)
    else:
      raise
  return os.fdopen(fd, "w")



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for working with python gRPC stubs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random
import time

import grpc

from tensorboard import version
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

# Default RPC timeout.
_GRPC_DEFAULT_TIMEOUT_SECS = 30

# Max number of times to attempt an RPC, retrying on transient failures.
_GRPC_RETRY_MAX_ATTEMPTS = 5

# Parameters to control the exponential backoff behavior.
_GRPC_RETRY_EXPONENTIAL_BASE = 2
_GRPC_RETRY_JITTER_FACTOR_MIN = 1.1
_GRPC_RETRY_JITTER_FACTOR_MAX = 1.5

# Status codes from gRPC for which it's reasonable to retry the RPC.
_GRPC_RETRYABLE_STATUS_CODES = frozenset([
    grpc.StatusCode.ABORTED,
    grpc.StatusCode.DEADLINE_EXCEEDED,
    grpc.StatusCode.RESOURCE_EXHAUSTED,
    grpc.StatusCode.UNAVAILABLE,
])

# gRPC metadata key whose value contains the client version.
_VERSION_METADATA_KEY = "tensorboard-version"


def call_with_retries(api_method, request, clock=None):
  """Call a gRPC stub API method, with automatic retry logic.

  This only supports unary-unary RPCs: i.e., no streaming on either end.
  Streamed RPCs will generally need application-level pagination support,
  because after a gRPC error one must retry the entire request; there is no
  "retry-resume" functionality.

  Args:
    api_method: Callable for the API method to invoke.
    request: Request protocol buffer to pass to the API method.
    clock: an interface object supporting `time()` and `sleep()` methods
      like the standard `time` module; if not passed, uses the normal module.

  Returns:
    Response protocol buffer returned by the API method.

  Raises:
    grpc.RpcError: if a non-retryable error is returned, or if all retry
      attempts have been exhausted.
  """
  if clock is None:
    clock = time
  # We can't actually use api_method.__name__ because it's not a real method,
  # it's a special gRPC callable instance that doesn't expose the method name.
  rpc_name = request.__class__.__name__.replace("Request", "")
  logger.debug("RPC call %s with request: %r", rpc_name, request)
  num_attempts = 0
  while True:
    num_attempts += 1
    try:
      return api_method(
          request,
          timeout=_GRPC_DEFAULT_TIMEOUT_SECS,
          metadata=version_metadata())
    except grpc.RpcError as e:
      logger.info("RPC call %s got error %s", rpc_name, e)
      if e.code() not in _GRPC_RETRYABLE_STATUS_CODES:
        raise
      if num_attempts >= _GRPC_RETRY_MAX_ATTEMPTS:
        raise
    jitter_factor = random.uniform(
        _GRPC_RETRY_JITTER_FACTOR_MIN, _GRPC_RETRY_JITTER_FACTOR_MAX)
    backoff_secs = (_GRPC_RETRY_EXPONENTIAL_BASE**num_attempts) * jitter_factor
    logger.info(
        "RPC call %s attempted %d times, retrying in %.1f seconds",
        rpc_name, num_attempts, backoff_secs)
    clock.sleep(backoff_secs)


def version_metadata():
  """Creates gRPC invocation metadata encoding the TensorBoard version.

  Usage: `stub.MyRpc(request, metadata=version_metadata())`.

  Returns:
    A tuple of key-value pairs (themselves 2-tuples) to be passed as the
    `metadata` kwarg to gRPC stub API methods.
  """
  return ((_VERSION_METADATA_KEY, version.VERSION),)


def extract_version(metadata):
  """Extracts version from invocation metadata.

  The argument should be the result of a prior call to `metadata` or the
  result of combining such a result with other metadata.

  Returns:
    The TensorBoard version listed in this metadata, or `None` if none
    is listed.
  """
  return dict(metadata).get(_VERSION_METADATA_KEY)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for use by the uploader command line tool."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import errno
import os
import os.path
import time


class RateLimiter(object):
  """Helper class for rate-limiting using a fixed minimum interval."""

  def __init__(self, interval_secs):
    """Constructs a RateLimiter that permits a tick() every `interval_secs`."""
    self._time = time  # Use property for ease of testing.
    self._interval_secs = interval_secs
    self._last_called_secs = 0

  def tick(self):
    """Blocks until it has been at least `interval_secs` since last tick()."""
    wait_secs = self._last_called_secs + self._interval_secs - self._time.time()
    if wait_secs > 0:
      self._time.sleep(wait_secs)
    self._last_called_secs = self._time.time()


def get_user_config_directory():
  """Returns a platform-specific root directory for user config settings."""
  # On Windows, prefer %LOCALAPPDATA%, then %APPDATA%, since we can expect the
  # AppData directories to be ACLed to be visible only to the user and admin
  # users (https://stackoverflow.com/a/7617601/1179226). If neither is set,
  # return None instead of falling back to something that may be world-readable.
  if os.name == "nt":
    appdata = os.getenv("LOCALAPPDATA")
    if appdata:
      return appdata
    appdata = os.getenv("APPDATA")
    if appdata:
      return appdata
    return None
  # On non-windows, use XDG_CONFIG_HOME if set, else default to ~/.config.
  xdg_config_home = os.getenv("XDG_CONFIG_HOME")
  if xdg_config_home:
    return xdg_config_home
  return os.path.join(os.path.expanduser("~"), ".config")


def make_file_with_directories(path, private=False):
  """Creates a file and its containing directories, if they don't already exist.


  If `private` is True, the file will be made private (readable only by the
  current user) and so will the leaf directory. Pre-existing contents of the
  file are not modified.

  Passing `private=True` is not supported on Windows because it doesn't support
  the relevant parts of `os.chmod()`.

  Args:
    path: str, The path of the file to create.
    private: boolean, Whether to make the file and leaf directory readable only
      by the current user.

  Raises:
    RuntimeError: If called on Windows with `private` set to True.
  """
  if private and os.name == "nt":
    raise RuntimeError("Creating private file not supported on Windows")
  try:
    path = os.path.realpath(path)
    leaf_dir = os.path.dirname(path)
    try:
      os.makedirs(leaf_dir)
    except OSError as e:
      if e.errno != errno.EEXIST:
        raise
    if private:
      os.chmod(leaf_dir, 0o700)
    open(path, "a").close()
    if private:
      os.chmod(path, 0o600)
  except EnvironmentError as e:
    raise RuntimeError("Failed to create file %s: %s" % (path, e))


def set_timestamp(pb, seconds_since_epoch):
  """Sets a `Timestamp` proto message to a floating point UNIX time.

  This is like `pb.FromNanoseconds(int(seconds_since_epoch * 1e9))` but
  without introducing floating-point error.

  Args:
    pb: A `google.protobuf.Timestamp` message to mutate.
    seconds_since_epoch: A `float`, as returned by `time.time`.
  """
  pb.seconds = int(seconds_since_epoch)
  pb.nanos = int(round((seconds_since_epoch % 1) * 10**9))


def format_time(timestamp_pb, now=None):
  """Converts a `timestamp_pb2.Timestamp` to human-readable string.

  This always includes the absolute date and time, and for recent dates
  may include a relative time like "(just now)" or "(2 hours ago)".

  Args:
    timestamp_pb: A `google.protobuf.timestamp_pb2.Timestamp` value to
      convert to string. The input will not be modified.
    now: A `datetime.datetime` object representing the current time,
      used for determining relative times like "just now". Optional;
      defaults to `datetime.datetime.now()`.

  Returns:
    A string suitable for human consumption.
  """

  # Add and subtract a day for <https://bugs.python.org/issue29097>,
  # which breaks early datetime conversions on Windows for small
  # timestamps.
  dt = datetime.datetime.fromtimestamp(timestamp_pb.seconds + 86400)
  dt = dt - datetime.timedelta(seconds=86400)

  if now is None:
    now = datetime.datetime.now()
  ago = now.replace(microsecond=0) - dt

  def ago_text(n, singular, plural):
    return "%d %s ago" % (n, singular if n == 1 else plural)

  relative = None
  if ago < datetime.timedelta(seconds=5):
    relative = "just now"
  elif ago < datetime.timedelta(minutes=1):
    relative = ago_text(int(ago.total_seconds()), "second", "seconds")
  elif ago < datetime.timedelta(hours=1):
    relative = ago_text(int(ago.total_seconds()) // 60, "minute", "minutes")
  elif ago < datetime.timedelta(days=1):
    relative = ago_text(int(ago.total_seconds()) // 3600, "hour", "hours")

  relative_part = " (%s)" % relative if relative is not None else ""
  return str(dt) + relative_part


def _ngettext(n, singular, plural):
  return "%d %s ago" % (n, singular if n == 1 else plural)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Mesh summaries and TensorFlow operations to create them. V2 versions"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json

from tensorboard.compat import tf2 as tf
from tensorboard.compat.proto import summary_pb2
from tensorboard.plugins.mesh import metadata
from tensorboard.plugins.mesh import plugin_data_pb2
from tensorboard.util import tensor_util


def _write_summary(
    name, description, tensor, content_type, components,
    json_config, step):
  """Creates a tensor summary with summary metadata.

  Args:
    name: A name for this summary. The summary tag used for TensorBoard will
      be this name prefixed by any active name scopes.
    description: Optional long-form description for this summary, as a
      constant `str`. Markdown is supported. Defaults to empty.
    tensor: Tensor to display in summary.
    content_type: Type of content inside the Tensor.
    components: Bitmask representing present parts (vertices, colors, etc.) that
      belong to the summary.
    json_config: A string, JSON-serialized dictionary of ThreeJS classes
      configuration.
    step: Explicit `int64`-castable monotonic step value for this summary. If
      omitted, this defaults to `tf.summary.experimental.get_step()`, which must
      not be None.

  Returns:
    A boolean indicating if summary was saved successfully or not.
  """
  tensor = tf.convert_to_tensor(value=tensor)
  shape = tensor.shape.as_list()
  shape = [dim if dim is not None else -1 for dim in shape]
  tensor_metadata = metadata.create_summary_metadata(
      name,
      None,  # display_name
      content_type,
      components,
      shape,
      description,
      json_config=json_config)
  return tf.summary.write(
        tag=metadata.get_instance_name(name, content_type),
        tensor=tensor,
        step=step,
        metadata=tensor_metadata)


def _get_json_config(config_dict):
  """Parses and returns JSON string from python dictionary."""
  json_config = '{}'
  if config_dict is not None:
    json_config = json.dumps(config_dict, sort_keys=True)
  return json_config


def mesh(name, vertices, faces=None, colors=None, config_dict=None, step=None,
   description=None):
  """Writes a TensorFlow mesh summary.

  Args:
    name: A name for this summary. The summary tag used for TensorBoard will
      be this name prefixed by any active name scopes.
    vertices: Tensor of shape `[dim_1, ..., dim_n, 3]` representing the 3D
      coordinates of vertices.
    faces: Tensor of shape `[dim_1, ..., dim_n, 3]` containing indices of
      vertices within each triangle.
    colors: Tensor of shape `[dim_1, ..., dim_n, 3]` containing colors for each
      vertex.
    config_dict: Dictionary with ThreeJS classes names and configuration.
    step: Explicit `int64`-castable monotonic step value for this summary. If
      omitted, this defaults to `tf.summary.experimental.get_step()`, which must
      not be None.
    description: Optional long-form description for this summary, as a
      constant `str`. Markdown is supported. Defaults to empty.

  Returns:
    True if all components of the mesh were saved successfully and False
      otherwise.
  """
  json_config = _get_json_config(config_dict)

  # All tensors representing a single mesh will be represented as separate
  # summaries internally. Those summaries will be regrouped on the client before
  # rendering.
  tensors = [
      metadata.MeshTensor(
        vertices, plugin_data_pb2.MeshPluginData.VERTEX, tf.float32),
      metadata.MeshTensor(faces, plugin_data_pb2.MeshPluginData.FACE, tf.int32),
      metadata.MeshTensor(
        colors, plugin_data_pb2.MeshPluginData.COLOR, tf.uint8)
  ]
  tensors = [tensor for tensor in tensors if tensor.data is not None]

  components = metadata.get_components_bitmask([
      tensor.content_type for tensor in tensors])

  summary_scope = (
      getattr(tf.summary.experimental, 'summary_scope', None) or
      tf.summary.summary_scope)
  all_success = True
  with summary_scope(name, 'mesh_summary', values=tensors):
    for tensor in tensors:
      all_success = all_success and _write_summary(
        name, description, tensor.data, tensor.content_type,
        components, json_config, step)

  return all_success


def mesh_pb(tag, vertices, faces=None, colors=None, config_dict=None,
   description=None):
  """Create a mesh summary to save in pb format.

  Args:
    tag: String tag for the summary.
    vertices: numpy array of shape `[dim_1, ..., dim_n, 3]` representing the 3D
      coordinates of vertices.
    faces: numpy array of shape `[dim_1, ..., dim_n, 3]` containing indices of
      vertices within each triangle.
    colors: numpy array of shape `[dim_1, ..., dim_n, 3]` containing colors for
      each vertex.
    config_dict: Dictionary with ThreeJS classes names and configuration.
    description: Optional long-form description for this summary, as a
      constant `str`. Markdown is supported. Defaults to empty.

  Returns:
    Instance of tf.Summary class.
  """
  json_config = _get_json_config(config_dict)

  summaries = []
  tensors = [
      metadata.MeshTensor(
          vertices, plugin_data_pb2.MeshPluginData.VERTEX, tf.float32),
      metadata.MeshTensor(faces, plugin_data_pb2.MeshPluginData.FACE, tf.int32),
      metadata.MeshTensor(
          colors, plugin_data_pb2.MeshPluginData.COLOR, tf.uint8)
  ]
  tensors = [tensor for tensor in tensors if tensor.data is not None]
  components = metadata.get_components_bitmask([
      tensor.content_type for tensor in tensors])
  for tensor in tensors:
    shape = tensor.data.shape
    shape = [dim if dim is not None else -1 for dim in shape]
    tensor_proto = tensor_util.make_tensor_proto(
        tensor.data, dtype=tensor.data_type)
    summary_metadata = metadata.create_summary_metadata(
        tag,
        None,  # display_name
        tensor.content_type,
        components,
        shape,
        description,
        json_config=json_config)
    instance_tag = metadata.get_instance_name(tag, tensor.content_type)
    summaries.append((instance_tag, summary_metadata, tensor_proto))

  summary = summary_pb2.Summary()
  for instance_tag, summary_metadata, tensor_proto in summaries:
    summary.value.add(
      tag=instance_tag, metadata=summary_metadata, tensor=tensor_proto)
  return summary



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Uploads a TensorBoard logdir to TensorBoard.dev."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import time

import grpc
import six

from tensorboard.uploader.proto import write_service_pb2
from tensorboard.uploader import logdir_loader
from tensorboard.uploader import peekable_iterator
from tensorboard.uploader import util
from tensorboard import data_compat
from tensorboard.backend.event_processing import directory_loader
from tensorboard.backend.event_processing import event_file_loader
from tensorboard.backend.event_processing import io_wrapper
from tensorboard.plugins.scalar import metadata as scalar_metadata
from tensorboard.util import grpc_util
from tensorboard.util import tb_logging
from tensorboard.util import tensor_util

# Minimum length of an upload cycle in seconds; shorter cycles will sleep to
# use up the rest of the time to avoid sending write RPCs too quickly.
_MIN_UPLOAD_CYCLE_DURATION_SECS = 5

# Age in seconds of last write after which an event file is considered inactive.
# TODO(@nfelt): consolidate with TensorBoard --reload_multifile default logic.
_EVENT_FILE_INACTIVE_SECS = 4000

# Maximum length of a base-128 varint as used to encode a 64-bit value
# (without the "msb of last byte is bit 63" optimization, to be
# compatible with protobuf and golang varints).
_MAX_VARINT64_LENGTH_BYTES = 10

# Maximum outgoing request size. The server-side limit is 4 MiB [1]; we
# should pad a bit to mitigate any errors in our bookkeeping. Currently,
# we pad a lot, because using higher request sizes causes occasional
# Deadline Exceeded errors in the RPC server.
#
# [1]: https://github.com/grpc/grpc/blob/e70d8582b4b0eedc45e3d25a57b58a08b94a9f4a/include/grpc/impl/codegen/grpc_types.h#L447  # pylint: disable=line-too-long
_MAX_REQUEST_LENGTH_BYTES = 1024 * 128

logger = tb_logging.get_logger()


class TensorBoardUploader(object):
  """Uploads a TensorBoard logdir to TensorBoard.dev."""

  def __init__(self, writer_client, logdir, rate_limiter=None):
    """Constructs a TensorBoardUploader.

    Args:
      writer_client: a TensorBoardWriterService stub instance
      logdir: path of the log directory to upload
      rate_limiter: a `RateLimiter` to use to limit upload cycle frequency
    """
    self._api = writer_client
    self._logdir = logdir
    self._request_builder = None
    if rate_limiter is None:
      self._rate_limiter = util.RateLimiter(_MIN_UPLOAD_CYCLE_DURATION_SECS)
    else:
      self._rate_limiter = rate_limiter
    active_filter = lambda secs: secs + _EVENT_FILE_INACTIVE_SECS >= time.time()
    directory_loader_factory = functools.partial(
        directory_loader.DirectoryLoader,
        loader_factory=event_file_loader.TimestampedEventFileLoader,
        path_filter=io_wrapper.IsTensorFlowEventsFile,
        active_filter=active_filter)
    self._logdir_loader = logdir_loader.LogdirLoader(
        self._logdir, directory_loader_factory)

  def create_experiment(self):
    """Creates an Experiment for this upload session and returns the ID."""
    logger.info("Creating experiment")
    request = write_service_pb2.CreateExperimentRequest()
    response = grpc_util.call_with_retries(self._api.CreateExperiment, request)
    self._request_builder = _RequestBuilder(response.experiment_id)
    return response.experiment_id

  def start_uploading(self):
    """Blocks forever to continuously upload data from the logdir.

    Raises:
      RuntimeError: If `create_experiment` has not yet been called.
      ExperimentNotFoundError: If the experiment is deleted during the
        course of the upload.
    """
    if self._request_builder is None:
      raise RuntimeError(
          "Must call create_experiment() before start_uploading()")
    while True:
      self._upload_once()

  def _upload_once(self):
    """Runs one upload cycle, sending zero or more RPCs."""
    logger.info("Starting an upload cycle")
    self._rate_limiter.tick()

    sync_start_time = time.time()
    self._logdir_loader.synchronize_runs()
    sync_duration_secs = time.time() - sync_start_time
    logger.info("Logdir sync took %.3f seconds", sync_duration_secs)

    run_to_events = self._logdir_loader.get_run_events()
    first_request = True
    for request in self._request_builder.build_requests(run_to_events):
      if not first_request:
        self._rate_limiter.tick()
      first_request = False
      upload_start_time = time.time()
      request_bytes = request.ByteSize()
      logger.info("Trying request of %d bytes", request_bytes)
      self._upload(request)
      upload_duration_secs = time.time() - upload_start_time
      logger.info(
          "Upload for %d runs (%d bytes) took %.3f seconds",
          len(request.runs),
          request_bytes,
          upload_duration_secs)

  def _upload(self, request):
    try:
      # TODO(@nfelt): execute this RPC asynchronously.
      grpc_util.call_with_retries(self._api.WriteScalar, request)
    except grpc.RpcError as e:
      if e.code() == grpc.StatusCode.NOT_FOUND:
        raise ExperimentNotFoundError()
      logger.error("Upload call failed with error %s", e)


def delete_experiment(writer_client, experiment_id):
  """Permanently deletes an experiment and all of its contents.

  Args:
    writer_client: a TensorBoardWriterService stub instance
    experiment_id: string ID of the experiment to delete

  Raises:
    ExperimentNotFoundError: If no such experiment exists.
    PermissionDeniedError: If the user is not authorized to delete this
      experiment.
    RuntimeError: On unexpected failure.
  """
  logger.info("Deleting experiment %r", experiment_id)
  request = write_service_pb2.DeleteExperimentRequest()
  request.experiment_id = experiment_id
  try:
    grpc_util.call_with_retries(writer_client.DeleteExperiment, request)
  except grpc.RpcError as e:
    if e.code() == grpc.StatusCode.NOT_FOUND:
      raise ExperimentNotFoundError()
    if e.code() == grpc.StatusCode.PERMISSION_DENIED:
      raise PermissionDeniedError()
    raise


class ExperimentNotFoundError(RuntimeError):
  pass


class PermissionDeniedError(RuntimeError):
  pass


class _OutOfSpaceError(Exception):
  """Action could not proceed without overflowing request budget.

  This is a signaling exception (like `StopIteration`) used internally
  by `_RequestBuilder`; it does not mean that anything has gone wrong.
  """
  pass


class _RequestBuilder(object):
  """Helper class for building requests that fit under a size limit.

  This class is not threadsafe. Use external synchronization if calling
  its methods concurrently.
  """

  _NON_SCALAR_TIME_SERIES = object()  # sentinel

  def __init__(self, experiment_id):
    self._experiment_id = experiment_id
    # The request currently being populated.
    self._request = None  # type: write_service_pb2.WriteScalarRequest
    # A lower bound on the number of bytes that we may yet add to the
    # request.
    self._byte_budget = None  # type: int
    # Map from `(run_name, tag_name)` to `SummaryMetadata` if the time
    # series is a scalar time series, else to `_NON_SCALAR_TIME_SERIES`.
    self._tag_metadata = {}

  def _new_request(self):
    """Allocates a new request and refreshes the budget."""
    self._request = write_service_pb2.WriteScalarRequest()
    self._byte_budget = _MAX_REQUEST_LENGTH_BYTES
    self._request.experiment_id = self._experiment_id
    self._byte_budget -= self._request.ByteSize()
    if self._byte_budget < 0:
      raise RuntimeError("Byte budget too small for experiment ID")

  def build_requests(self, run_to_events):
    """Converts a stream of TF events to a stream of outgoing requests.

    Each yielded request will be at most `_MAX_REQUEST_LENGTH_BYTES`
    bytes long.

    Args:
      run_to_events: Mapping from run name to generator of `tf.Event`
        values, as returned by `LogdirLoader.get_run_events`.

    Yields:
      A finite stream of `WriteScalarRequest` objects.

    Raises:
      RuntimeError: If no progress can be made because even a single
      point is too large (say, due to a gigabyte-long tag name).
    """

    self._new_request()
    runs = {}  # cache: map from run name to `Run` proto in request
    tags = {}  # cache: map from `(run, tag)` to `Tag` proto in run in request
    work_items = peekable_iterator.PeekableIterator(
        self._run_values(run_to_events))

    while work_items.has_next():
      (run_name, event, orig_value) = work_items.peek()
      value = data_compat.migrate_value(orig_value)
      time_series_key = (run_name, value.tag)

      metadata = self._tag_metadata.get(time_series_key)
      if metadata is None:
        plugin_name = value.metadata.plugin_data.plugin_name
        if plugin_name == scalar_metadata.PLUGIN_NAME:
          metadata = value.metadata
        else:
          metadata = _RequestBuilder._NON_SCALAR_TIME_SERIES
        self._tag_metadata[time_series_key] = metadata
      if metadata is _RequestBuilder._NON_SCALAR_TIME_SERIES:
        next(work_items)
        continue
      try:
        run_proto = runs.get(run_name)
        if run_proto is None:
          run_proto = self._create_run(run_name)
          runs[run_name] = run_proto
        tag_proto = tags.get((run_name, value.tag))
        if tag_proto is None:
          tag_proto = self._create_tag(run_proto, value.tag, metadata)
          tags[(run_name, value.tag)] = tag_proto
        self._create_point(tag_proto, event, value)
        next(work_items)
      except _OutOfSpaceError:
        # Flush request and start a new one.
        request_to_emit = self._prune_request()
        if request_to_emit is None:
          raise RuntimeError("Could not make progress uploading data")
        self._new_request()
        runs.clear()
        tags.clear()
        yield request_to_emit

    final_request = self._prune_request()
    if final_request is not None:
      yield final_request

  def _run_values(self, run_to_events):
    """Helper generator to create a single stream of work items."""
    # Note that each of these joins in principle has deletion anomalies:
    # if the input stream contains runs with no events, or events with
    # no values, we'll lose that information. This is not a problem: we
    # would need to prune such data from the request anyway.
    for (run_name, events) in six.iteritems(run_to_events):
      for event in events:
        for value in event.summary.value:
          yield (run_name, event, value)

  def _prune_request(self):
    """Removes empty runs and tags from the active request.

    This does not refund `self._byte_budget`; it is assumed that the
    request will be emitted immediately, anyway.

    Returns:
      The active request, or `None` if after pruning the request
      contains no data.
    """
    request = self._request
    for (run_idx, run) in reversed(list(enumerate(request.runs))):
      for (tag_idx, tag) in reversed(list(enumerate(run.tags))):
        if not tag.points:
          del run.tags[tag_idx]
      if not run.tags:
        del self._request.runs[run_idx]
    if not request.runs:
      request = None
    return request

  def _create_run(self, run_name):
    """Adds a run to the live request, if there's space.

    Args:
      run_name: String name of the run to add.

    Returns:
      The `WriteScalarRequest.Run` that was added to `request.runs`.

    Raises:
      _OutOfSpaceError: If adding the run would exceed the remaining
        request budget.
    """
    run_proto = self._request.runs.add(name=run_name)
    # We can't calculate the proto key cost exactly ahead of time, as
    # it depends on the total size of all tags. Be conservative.
    cost = run_proto.ByteSize() + _MAX_VARINT64_LENGTH_BYTES + 1
    if cost > self._byte_budget:
      raise _OutOfSpaceError()
    self._byte_budget -= cost
    return run_proto

  def _create_tag(self, run_proto, tag_name, metadata):
    """Adds a tag for the given value, if there's space.

    Args:
      run_proto: `WriteScalarRequest.Run` proto to which to add a tag.
      tag_name: String name of the tag to add (as `value.tag`).
      metadata: TensorBoard `SummaryMetadata` proto from the first
        occurrence of this time series.

    Returns:
      The `WriteScalarRequest.Tag` that was added to `run_proto.tags`.

    Raises:
      _OutOfSpaceError: If adding the tag would exceed the remaining
        request budget.
    """
    tag_proto = run_proto.tags.add(name=tag_name)
    tag_proto.metadata.CopyFrom(metadata)
    submessage_cost = tag_proto.ByteSize()
    # We can't calculate the proto key cost exactly ahead of time, as
    # it depends on the number of points. Be conservative.
    cost = submessage_cost + _MAX_VARINT64_LENGTH_BYTES + 1
    if cost > self._byte_budget:
      raise _OutOfSpaceError()
    self._byte_budget -= cost
    return tag_proto

  def _create_point(self, tag_proto, event, value):
    """Adds a scalar point to the given tag, if there's space.

    Args:
      tag_proto: `WriteScalarRequest.Tag` proto to which to add a point.
      event: Enclosing `Event` proto with the step and wall time data.
      value: Scalar `Summary.Value` proto with the actual scalar data.

    Returns:
      The `ScalarPoint` that was added to `tag_proto.points`.

    Raises:
      _OutOfSpaceError: If adding the point would exceed the remaining
        request budget.
    """
    point = tag_proto.points.add()
    point.step = event.step
    # TODO(@nfelt): skip tensor roundtrip for Value with simple_value set
    point.value = tensor_util.make_ndarray(value.tensor).item()
    util.set_timestamp(point.wall_time, event.wall_time)
    submessage_cost = point.ByteSize()
    cost = submessage_cost + _varint_cost(submessage_cost) + 1  # proto key
    if cost > self._byte_budget:
      tag_proto.points.pop()
      raise _OutOfSpaceError()
    self._byte_budget -= cost
    return point


def _varint_cost(n):
  """Computes the size of `n` encoded as an unsigned base-128 varint.

  This should be consistent with the proto wire format:
  <https://developers.google.com/protocol-buffers/docs/encoding#varints>

  Args:
    n: A non-negative integer.

  Returns:
    An integer number of bytes.
  """
  result = 1
  while n >= 128:
    result += 1
    n >>= 7
  return result



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Collection of first-party plugins.

This module exists to isolate tensorboard.program from the potentially
heavyweight build dependencies for first-party plugins. This way people
doing custom builds of TensorBoard have the option to only pay for the
dependencies they want.

This module also grants the flexibility to those doing custom builds, to
automatically inherit the centrally-maintained list of standard plugins,
for less repetition.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os

import pkg_resources

from tensorboard.compat import tf
from tensorboard.plugins import base_plugin
from tensorboard.plugins.audio import audio_plugin
from tensorboard.plugins.beholder import beholder_plugin_loader
from tensorboard.plugins.core import core_plugin
from tensorboard.plugins.custom_scalar import custom_scalars_plugin
from tensorboard.plugins.debugger import debugger_plugin_loader
from tensorboard.plugins.distribution import distributions_plugin
from tensorboard.plugins.graph import graphs_plugin
from tensorboard.plugins.histogram import histograms_plugin
from tensorboard.plugins.hparams import hparams_plugin
from tensorboard.plugins.image import images_plugin
from tensorboard.plugins.interactive_inference import (
    interactive_inference_plugin_loader
)
from tensorboard.plugins.pr_curve import pr_curves_plugin
from tensorboard.plugins.profile import profile_plugin_loader
from tensorboard.plugins.scalar import scalars_plugin
from tensorboard.plugins.text import text_plugin
from tensorboard.plugins.mesh import mesh_plugin


logger = logging.getLogger(__name__)

# Ordering matters. The order in which these lines appear determines the
# ordering of tabs in TensorBoard's GUI.
_PLUGINS = [
    core_plugin.CorePluginLoader,
    scalars_plugin.ScalarsPlugin,
    custom_scalars_plugin.CustomScalarsPlugin,
    images_plugin.ImagesPlugin,
    audio_plugin.AudioPlugin,
    debugger_plugin_loader.DebuggerPluginLoader,
    graphs_plugin.GraphsPlugin,
    distributions_plugin.DistributionsPlugin,
    histograms_plugin.HistogramsPlugin,
    text_plugin.TextPlugin,
    pr_curves_plugin.PrCurvesPlugin,
    profile_plugin_loader.ProfilePluginLoader,
    beholder_plugin_loader.BeholderPluginLoader,
    interactive_inference_plugin_loader.InteractiveInferencePluginLoader,
    hparams_plugin.HParamsPlugin,
    mesh_plugin.MeshPlugin,
]

def get_plugins():
  """Returns a list specifying TensorBoard's default first-party plugins.

  Plugins are specified in this list either via a TBLoader instance to load the
  plugin, or the TBPlugin class itself which will be loaded using a BasicLoader.

  This list can be passed to the `tensorboard.program.TensorBoard` API.

  Returns:
    The list of default plugins.

  :rtype: list[Type[base_plugin.TBLoader] | Type[base_plugin.TBPlugin]]
  """

  return _PLUGINS[:]


def get_dynamic_plugins():
  """Returns a list specifying TensorBoard's dynamically loaded plugins.

  A dynamic TensorBoard plugin is specified using entry_points [1] and it is
  the robust way to integrate plugins into TensorBoard.

  This list can be passed to the `tensorboard.program.TensorBoard` API.

  Returns:
    The list of dynamic plugins.

  :rtype: list[Type[base_plugin.TBLoader] | Type[base_plugin.TBPlugin]]

  [1]: https://packaging.python.org/specifications/entry-points/
  """
  return [
      entry_point.load()
      for entry_point in pkg_resources.iter_entry_points('tensorboard_plugins')
  ]



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for TensorBoard command line program.

This is a lightweight module for bringing up a TensorBoard HTTP server
or emulating the `tensorboard` shell command.

Those wishing to create custom builds of TensorBoard can use this module
by swapping out `tensorboard.main` with the custom definition that
modifies the set of plugins and static assets.

This module does not depend on first-party plugins or the default web
server assets. Those are defined in `tensorboard.default`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABCMeta
from abc import abstractmethod
import argparse
import atexit
from collections import defaultdict
import errno
import inspect
import logging
import os
import signal
import socket
import sys
import threading
import time

from absl import flags as absl_flags
from absl.flags import argparse_flags
import absl.logging
import six
from six.moves import urllib
from six.moves import xrange  # pylint: disable=redefined-builtin
from werkzeug import serving

from tensorboard import manager
from tensorboard import version
from tensorboard.backend import application
from tensorboard.backend.event_processing import event_file_inspector as efi
from tensorboard.plugins import base_plugin
from tensorboard.plugins.core import core_plugin
from tensorboard.util import argparse_util
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()

# Default subcommand name. This is a user-facing CLI and should not change.
_SERVE_SUBCOMMAND_NAME = 'serve'
# Internal flag name used to store which subcommand was invoked.
_SUBCOMMAND_FLAG = '__tensorboard_subcommand'


def setup_environment():
  """Makes recommended modifications to the environment.

  This functions changes global state in the Python process. Calling
  this function is a good idea, but it can't appropriately be called
  from library routines.
  """
  absl.logging.set_verbosity(absl.logging.WARNING)

  # The default is HTTP/1.0 for some strange reason. If we don't use
  # HTTP/1.1 then a new TCP socket and Python thread is created for
  # each HTTP request. The tradeoff is we must always specify the
  # Content-Length header, or do chunked encoding for streaming.
  serving.WSGIRequestHandler.protocol_version = 'HTTP/1.1'

def get_default_assets_zip_provider():
  """Opens stock TensorBoard web assets collection.

  Returns:
    Returns function that returns a newly opened file handle to zip file
    containing static assets for stock TensorBoard, or None if webfiles.zip
    could not be found. The value the callback returns must be closed. The
    paths inside the zip file are considered absolute paths on the web server.
  """
  path = os.path.join(os.path.dirname(inspect.getfile(sys._getframe(1))),
                      'webfiles.zip')
  if not os.path.exists(path):
    logger.warning('webfiles.zip static assets not found: %s', path)
    return None
  return lambda: open(path, 'rb')

class TensorBoard(object):
  """Class for running TensorBoard.

  Fields:
    plugin_loaders: Set from plugins passed to constructor.
    assets_zip_provider: Set by constructor.
    server_class: Set by constructor.
    flags: An argparse.Namespace set by the configure() method.
    cache_key: As `manager.cache_key`; set by the configure() method.
  """

  def __init__(
      self,
      plugins=None,
      assets_zip_provider=None,
      server_class=None,
      subcommands=None,
  ):
    """Creates new instance.

    Args:
      plugins: A list of TensorBoard plugins to load, as TBPlugin classes or
        TBLoader instances or classes. If not specified, defaults to first-party
        plugins.
      assets_zip_provider: Delegates to TBContext or uses default if None.
      server_class: An optional factory for a `TensorBoardServer` to use
        for serving the TensorBoard WSGI app. If provided, its callable
        signature should match that of `TensorBoardServer.__init__`.

    :type plugins:
      list[
        base_plugin.TBLoader | Type[base_plugin.TBLoader] |
        Type[base_plugin.TBPlugin]
      ]
    """
    if plugins is None:
      from tensorboard import default
      plugins = default.get_plugins()
    if assets_zip_provider is None:
      assets_zip_provider = get_default_assets_zip_provider()
    if server_class is None:
      server_class = create_port_scanning_werkzeug_server
    if subcommands is None:
      subcommands = []
    self.plugin_loaders = [application.make_plugin_loader(p) for p in plugins]
    self.assets_zip_provider = assets_zip_provider
    self.server_class = server_class
    self.subcommands = {}
    for subcommand in subcommands:
      name = subcommand.name()
      if name in self.subcommands or name == _SERVE_SUBCOMMAND_NAME:
        raise ValueError("Duplicate subcommand name: %r" % name)
      self.subcommands[name] = subcommand
    self.flags = None

  def configure(self, argv=('',), **kwargs):
    """Configures TensorBoard behavior via flags.

    This method will populate the "flags" property with an argparse.Namespace
    representing flag values parsed from the provided argv list, overridden by
    explicit flags from remaining keyword arguments.

    Args:
      argv: Can be set to CLI args equivalent to sys.argv; the first arg is
        taken to be the name of the path being executed.
      kwargs: Additional arguments will override what was parsed from
        argv. They must be passed as Python data structures, e.g.
        `foo=1` rather than `foo="1"`.

    Returns:
      Either argv[:1] if argv was non-empty, or [''] otherwise, as a mechanism
      for absl.app.run() compatibility.

    Raises:
      ValueError: If flag values are invalid.
    """

    base_parser = argparse_flags.ArgumentParser(
        prog='tensorboard',
        description=('TensorBoard is a suite of web applications for '
                     'inspecting and understanding your TensorFlow runs '
                     'and graphs. https://github.com/tensorflow/tensorboard '))
    subparsers = base_parser.add_subparsers(
        help="TensorBoard subcommand (defaults to %r)" % _SERVE_SUBCOMMAND_NAME)

    serve_subparser = subparsers.add_parser(
        _SERVE_SUBCOMMAND_NAME,
        help='start local TensorBoard server (default subcommand)')
    serve_subparser.set_defaults(**{_SUBCOMMAND_FLAG: _SERVE_SUBCOMMAND_NAME})

    if len(argv) < 2 or argv[1].startswith('-'):
      # This invocation, if valid, must not use any subcommands: we
      # don't permit flags before the subcommand name.
      serve_parser = base_parser
    else:
      # This invocation, if valid, must use a subcommand: we don't take
      # any positional arguments to `serve`.
      serve_parser = serve_subparser

    for (name, subcommand) in six.iteritems(self.subcommands):
      subparser = subparsers.add_parser(
          name, help=subcommand.help(), description=subcommand.description())
      subparser.set_defaults(**{_SUBCOMMAND_FLAG: name})
      subcommand.define_flags(subparser)

    for loader in self.plugin_loaders:
      loader.define_flags(serve_parser)

    arg0 = argv[0] if argv else ''

    with argparse_util.allow_missing_subcommand():
      flags = base_parser.parse_args(argv[1:])  # Strip binary name from argv.
    if getattr(flags, _SUBCOMMAND_FLAG, None) is None:
      # Manually assign default value rather than using `set_defaults`
      # on the base parser to work around Python bug #9351 on old
      # versions of `argparse`: <https://bugs.python.org/issue9351>
      setattr(flags, _SUBCOMMAND_FLAG, _SERVE_SUBCOMMAND_NAME)

    self.cache_key = manager.cache_key(
        working_directory=os.getcwd(),
        arguments=argv[1:],
        configure_kwargs=kwargs,
    )
    if arg0:
      # Only expose main module Abseil flags as TensorBoard native flags.
      # This is the same logic Abseil's ArgumentParser uses for determining
      # which Abseil flags to include in the short helpstring.
      for flag in set(absl_flags.FLAGS.get_key_flags_for_module(arg0)):
        if hasattr(flags, flag.name):
          raise ValueError('Conflicting Abseil flag: %s' % flag.name)
        setattr(flags, flag.name, flag.value)
    for k, v in kwargs.items():
      if not hasattr(flags, k):
        raise ValueError('Unknown TensorBoard flag: %s' % k)
      setattr(flags, k, v)
    if getattr(flags, _SUBCOMMAND_FLAG) == _SERVE_SUBCOMMAND_NAME:
      for loader in self.plugin_loaders:
        loader.fix_flags(flags)
    self.flags = flags
    return [arg0]

  def main(self, ignored_argv=('',)):
    """Blocking main function for TensorBoard.

    This method is called by `tensorboard.main.run_main`, which is the
    standard entrypoint for the tensorboard command line program. The
    configure() method must be called first.

    Args:
      ignored_argv: Do not pass. Required for Abseil compatibility.

    Returns:
      Process exit code, i.e. 0 if successful or non-zero on failure. In
      practice, an exception will most likely be raised instead of
      returning non-zero.

    :rtype: int
    """
    self._install_signal_handler(signal.SIGTERM, "SIGTERM")
    subcommand_name = getattr(self.flags, _SUBCOMMAND_FLAG)
    if subcommand_name == _SERVE_SUBCOMMAND_NAME:
      runner = self._run_serve_subcommand
    else:
      runner = self.subcommands[subcommand_name].run
    return runner(self.flags) or 0

  def _run_serve_subcommand(self, flags):
    # TODO(#2801): Make `--version` a flag on only the base parser, not `serve`.
    if flags.version_tb:
      print(version.VERSION)
      return 0
    if flags.inspect:
      # TODO(@wchargin): Convert `inspect` to a normal subcommand?
      logger.info('Not bringing up TensorBoard, but inspecting event files.')
      event_file = os.path.expanduser(flags.event_file)
      efi.inspect(flags.logdir, event_file, flags.tag)
      return 0
    try:
      server = self._make_server()
      server.print_serving_message()
      self._register_info(server)
      server.serve_forever()
      return 0
    except TensorBoardServerException as e:
      logger.error(e.msg)
      sys.stderr.write('ERROR: %s\n' % e.msg)
      sys.stderr.flush()
      return -1

  def launch(self):
    """Python API for launching TensorBoard.

    This method is the same as main() except it launches TensorBoard in
    a separate permanent thread. The configure() method must be called
    first.

    Returns:
      The URL of the TensorBoard web server.

    :rtype: str
    """
    # Make it easy to run TensorBoard inside other programs, e.g. Colab.
    server = self._make_server()
    thread = threading.Thread(target=server.serve_forever, name='TensorBoard')
    thread.daemon = True
    thread.start()
    return server.get_url()

  def _register_info(self, server):
    """Write a TensorBoardInfo file and arrange for its cleanup.

    Args:
      server: The result of `self._make_server()`.
    """
    server_url = urllib.parse.urlparse(server.get_url())
    info = manager.TensorBoardInfo(
        version=version.VERSION,
        start_time=int(time.time()),
        port=server_url.port,
        pid=os.getpid(),
        path_prefix=self.flags.path_prefix,
        logdir=self.flags.logdir or self.flags.logdir_spec,
        db=self.flags.db,
        cache_key=self.cache_key,
    )
    atexit.register(manager.remove_info_file)
    manager.write_info_file(info)

  def _install_signal_handler(self, signal_number, signal_name):
    """Set a signal handler to gracefully exit on the given signal.

    When this process receives the given signal, it will run `atexit`
    handlers and then exit with `0`.

    Args:
      signal_number: The numeric code for the signal to handle, like
        `signal.SIGTERM`.
      signal_name: The human-readable signal name.
    """
    old_signal_handler = None  # set below
    def handler(handled_signal_number, frame):
      # In case we catch this signal again while running atexit
      # handlers, take the hint and actually die.
      signal.signal(signal_number, signal.SIG_DFL)
      sys.stderr.write("TensorBoard caught %s; exiting...\n" % signal_name)
      # The main thread is the only non-daemon thread, so it suffices to
      # exit hence.
      if old_signal_handler not in (signal.SIG_IGN, signal.SIG_DFL):
        old_signal_handler(handled_signal_number, frame)
      sys.exit(0)
    old_signal_handler = signal.signal(signal_number, handler)


  def _make_server(self):
    """Constructs the TensorBoard WSGI app and instantiates the server."""
    app = application.standard_tensorboard_wsgi(self.flags,
                                                self.plugin_loaders,
                                                self.assets_zip_provider)
    return self.server_class(app, self.flags)


@six.add_metaclass(ABCMeta)
class TensorBoardSubcommand(object):
  """Experimental private API for defining subcommands to tensorboard(1)."""

  @abstractmethod
  def name(self):
    """Name of this subcommand, as specified on the command line.

    This must be unique across all subcommands.

    Returns:
      A string.
    """
    pass

  @abstractmethod
  def define_flags(self, parser):
    """Configure an argument parser for this subcommand.

    Flags whose names start with two underscores (e.g., `__foo`) are
    reserved for use by the runtime and must not be defined by
    subcommands.

    Args:
      parser: An `argparse.ArgumentParser` scoped to this subcommand,
        which this function should mutate.
    """
    pass

  @abstractmethod
  def run(self, flags):
    """Execute this subcommand with user-provided flags.

    Args:
      flags: An `argparse.Namespace` object with all defined flags.

    Returns:
      An `int` exit code, or `None` as an alias for `0`.
    """
    pass

  def help(self):
    """Short, one-line help text to display on `tensorboard --help`."""
    return None

  def description(self):
    """Description to display on `tensorboard SUBCOMMAND --help`."""
    return None


@six.add_metaclass(ABCMeta)
class TensorBoardServer(object):
  """Class for customizing TensorBoard WSGI app serving."""

  @abstractmethod
  def __init__(self, wsgi_app, flags):
    """Create a flag-configured HTTP server for TensorBoard's WSGI app.

    Args:
      wsgi_app: The TensorBoard WSGI application to create a server for.
      flags: argparse.Namespace instance of TensorBoard flags.
    """
    raise NotImplementedError()

  @abstractmethod
  def serve_forever(self):
    """Blocking call to start serving the TensorBoard server."""
    raise NotImplementedError()

  @abstractmethod
  def get_url(self):
    """Returns a URL at which this server should be reachable."""
    raise NotImplementedError()

  def print_serving_message(self):
    """Prints a user-friendly message prior to server start.

    This will be called just before `serve_forever`.
    """
    sys.stderr.write(
        'TensorBoard %s at %s (Press CTRL+C to quit)\n'
        % (version.VERSION, self.get_url())
    )
    sys.stderr.flush()


class TensorBoardServerException(Exception):
  """Exception raised by TensorBoardServer for user-friendly errors.

  Subclasses of TensorBoardServer can raise this exception in order to
  generate a clean error message for the user rather than a stacktrace.
  """
  def __init__(self, msg):
    self.msg = msg


class TensorBoardPortInUseError(TensorBoardServerException):
  """Error raised when attempting to bind to a port that is in use.

  This should be raised when it is expected that binding to another
  similar port would succeed. It is used as a signal to indicate that
  automatic port searching should continue rather than abort.
  """
  pass


def with_port_scanning(cls):
  """Create a server factory that performs port scanning.

  This function returns a callable whose signature matches the
  specification of `TensorBoardServer.__init__`, using `cls` as an
  underlying implementation. It passes through `flags` unchanged except
  in the case that `flags.port is None`, in which case it repeatedly
  instantiates the underlying server with new port suggestions.

  Args:
    cls: A valid implementation of `TensorBoardServer`. This class's
      initializer should raise a `TensorBoardPortInUseError` upon
      failing to bind to a port when it is expected that binding to
      another nearby port might succeed.

      The initializer for `cls` will only ever be invoked with `flags`
      such that `flags.port is not None`.

  Returns:
    A function that implements the `__init__` contract of
    `TensorBoardServer`.
  """

  def init(wsgi_app, flags):
    # base_port: what's the first port to which we should try to bind?
    # should_scan: if that fails, shall we try additional ports?
    # max_attempts: how many ports shall we try?
    should_scan = flags.port is None
    base_port = core_plugin.DEFAULT_PORT if flags.port is None else flags.port
    max_attempts = 10 if should_scan else 1

    if base_port > 0xFFFF:
      raise TensorBoardServerException(
          'TensorBoard cannot bind to port %d > %d' % (base_port, 0xFFFF)
      )
    max_attempts = 10 if should_scan else 1
    base_port = min(base_port + max_attempts, 0x10000) - max_attempts

    for port in xrange(base_port, base_port + max_attempts):
      subflags = argparse.Namespace(**vars(flags))
      subflags.port = port
      try:
        return cls(wsgi_app=wsgi_app, flags=subflags)
      except TensorBoardPortInUseError:
        if not should_scan:
          raise
    # All attempts failed to bind.
    raise TensorBoardServerException(
        'TensorBoard could not bind to any port around %s '
        '(tried %d times)'
        % (base_port, max_attempts))

  return init


class WerkzeugServer(serving.ThreadedWSGIServer, TensorBoardServer):
  """Implementation of TensorBoardServer using the Werkzeug dev server."""

  # ThreadedWSGIServer handles this in werkzeug 0.12+ but we allow 0.11.x.
  daemon_threads = True

  def __init__(self, wsgi_app, flags):
    self._flags = flags
    host = flags.host
    port = flags.port

    self._auto_wildcard = flags.bind_all
    if self._auto_wildcard:
      # Serve on all interfaces, and attempt to serve both IPv4 and IPv6
      # traffic through one socket.
      host = self._get_wildcard_address(port)
    elif host is None:
      host = 'localhost'

    self._host = host

    self._fix_werkzeug_logging()
    try:
      super(WerkzeugServer, self).__init__(host, port, wsgi_app)
    except socket.error as e:
      if hasattr(errno, 'EACCES') and e.errno == errno.EACCES:
        raise TensorBoardServerException(
            'TensorBoard must be run as superuser to bind to port %d' %
            port)
      elif hasattr(errno, 'EADDRINUSE') and e.errno == errno.EADDRINUSE:
        if port == 0:
          raise TensorBoardServerException(
              'TensorBoard unable to find any open port')
        else:
          raise TensorBoardPortInUseError(
              'TensorBoard could not bind to port %d, it was already in use' %
              port)
      elif hasattr(errno, 'EADDRNOTAVAIL') and e.errno == errno.EADDRNOTAVAIL:
        raise TensorBoardServerException(
            'TensorBoard could not bind to unavailable address %s' % host)
      elif hasattr(errno, 'EAFNOSUPPORT') and e.errno == errno.EAFNOSUPPORT:
        raise TensorBoardServerException(
            'Tensorboard could not bind to unsupported address family %s' %
            host)
      # Raise the raw exception if it wasn't identifiable as a user error.
      raise

  def _get_wildcard_address(self, port):
    """Returns a wildcard address for the port in question.

    This will attempt to follow the best practice of calling getaddrinfo() with
    a null host and AI_PASSIVE to request a server-side socket wildcard address.
    If that succeeds, this returns the first IPv6 address found, or if none,
    then returns the first IPv4 address. If that fails, then this returns the
    hardcoded address "::" if socket.has_ipv6 is True, else "0.0.0.0".
    """
    fallback_address = '::' if socket.has_ipv6 else '0.0.0.0'
    if hasattr(socket, 'AI_PASSIVE'):
      try:
        addrinfos = socket.getaddrinfo(None, port, socket.AF_UNSPEC,
                                       socket.SOCK_STREAM, socket.IPPROTO_TCP,
                                       socket.AI_PASSIVE)
      except socket.gaierror as e:
        logger.warn('Failed to auto-detect wildcard address, assuming %s: %s',
                    fallback_address, str(e))
        return fallback_address
      addrs_by_family = defaultdict(list)
      for family, _, _, _, sockaddr in addrinfos:
        # Format of the "sockaddr" socket address varies by address family,
        # but [0] is always the IP address portion.
        addrs_by_family[family].append(sockaddr[0])
      if hasattr(socket, 'AF_INET6') and addrs_by_family[socket.AF_INET6]:
        return addrs_by_family[socket.AF_INET6][0]
      if hasattr(socket, 'AF_INET') and addrs_by_family[socket.AF_INET]:
        return addrs_by_family[socket.AF_INET][0]
    logger.warn('Failed to auto-detect wildcard address, assuming %s',
                fallback_address)
    return fallback_address

  def server_bind(self):
    """Override to enable IPV4 mapping for IPV6 sockets when desired.

    The main use case for this is so that when no host is specified, TensorBoard
    can listen on all interfaces for both IPv4 and IPv6 connections, rather than
    having to choose v4 or v6 and hope the browser didn't choose the other one.
    """
    socket_is_v6 = (
        hasattr(socket, 'AF_INET6') and self.socket.family == socket.AF_INET6)
    has_v6only_option = (
        hasattr(socket, 'IPPROTO_IPV6') and hasattr(socket, 'IPV6_V6ONLY'))
    if self._auto_wildcard and socket_is_v6 and has_v6only_option:
      try:
        self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
      except socket.error as e:
        # Log a warning on failure to dual-bind, except for EAFNOSUPPORT
        # since that's expected if IPv4 isn't supported at all (IPv6-only).
        if hasattr(errno, 'EAFNOSUPPORT') and e.errno != errno.EAFNOSUPPORT:
          logger.warn('Failed to dual-bind to IPv4 wildcard: %s', str(e))
    super(WerkzeugServer, self).server_bind()

  def handle_error(self, request, client_address):
    """Override to get rid of noisy EPIPE errors."""
    del request  # unused
    # Kludge to override a SocketServer.py method so we can get rid of noisy
    # EPIPE errors. They're kind of a red herring as far as errors go. For
    # example, `curl -N http://localhost:6006/ | head` will cause an EPIPE.
    exc_info = sys.exc_info()
    e = exc_info[1]
    if isinstance(e, IOError) and e.errno == errno.EPIPE:
      logger.warn('EPIPE caused by %s in HTTP serving' % str(client_address))
    else:
      logger.error('HTTP serving error', exc_info=exc_info)

  def get_url(self):
    if self._auto_wildcard:
      display_host = socket.gethostname()
    else:
      host = self._host
      display_host = (
          '[%s]' % host if ':' in host and not host.startswith('[') else host)
    return 'http://%s:%d%s/' % (display_host, self.server_port,
                               self._flags.path_prefix.rstrip('/'))

  def print_serving_message(self):
    if self._flags.host is None and not self._flags.bind_all:
      sys.stderr.write(
          'Serving TensorBoard on localhost; to expose to the network, '
          'use a proxy or pass --bind_all\n'
      )
      sys.stderr.flush()
    super(WerkzeugServer, self).print_serving_message()

  def _fix_werkzeug_logging(self):
    """Fix werkzeug logging setup so it inherits TensorBoard's log level.

    This addresses a change in werkzeug 0.15.0+ [1] that causes it set its own
    log level to INFO regardless of the root logger configuration. We instead
    want werkzeug to inherit TensorBoard's root logger log level (set via absl
    to WARNING by default).

    [1]: https://github.com/pallets/werkzeug/commit/4cf77d25858ff46ac7e9d64ade054bf05b41ce12
    """
    # Log once at DEBUG to force werkzeug to initialize its singleton logger,
    # which sets the logger level to INFO it if is unset, and then access that
    # object via logging.getLogger('werkzeug') to durably revert the level to
    # unset (and thus make messages logged to it inherit the root logger level).
    self.log('debug', 'Fixing werkzeug logger to inherit TensorBoard log level')
    logging.getLogger('werkzeug').setLevel(logging.NOTSET)


create_port_scanning_werkzeug_server = with_port_scanning(WerkzeugServer)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines an error (exception) class for the HParams plugin."""


class HParamsError(Exception):
  """Represents an error that is meaningful to the end-user. Such an error
  should have a meaningful error message. Other errors, (such as resulting
  from some internal invariants being violated) should be represented by
  other exceptions.
  """
  pass



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Loader for event file data for an entire TensorBoard log directory."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os

from tensorboard.backend.event_processing import directory_watcher
from tensorboard.backend.event_processing import io_wrapper
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()


class LogdirLoader(object):
  """Loader for a root log directory, maintaining multiple DirectoryLoaders.

  This class takes a root log directory and a factory for DirectoryLoaders, and
  maintains one DirectoryLoader per "logdir subdirectory" of the root logdir.

  Note that this class is not thread-safe.
  """

  def __init__(self, logdir, directory_loader_factory):
    """Constructs a new LogdirLoader.

    Args:
      logdir: The root log directory to load from.
      directory_loader_factory: A factory for creating DirectoryLoaders. The
        factory should take a path and return a DirectoryLoader.

    Raises:
      ValueError: If logdir or directory_loader_factory are None.
    """
    if logdir is None:
      raise ValueError('A logdir is required')
    if directory_loader_factory is None:
      raise ValueError('A directory loader factory is required')
    self._logdir = logdir
    self._directory_loader_factory = directory_loader_factory
    # Maps run names to corresponding DirectoryLoader instances.
    self._directory_loaders = {}

  def synchronize_runs(self):
    """Finds new runs within `logdir` and makes `DirectoryLoaders` for them.

    In addition, any existing `DirectoryLoader` whose run directory no longer
    exists will be deleted.
    """
    logger.info('Starting logdir traversal of %s', self._logdir)
    runs_seen = set()
    for subdir in io_wrapper.GetLogdirSubdirectories(self._logdir):
      run = os.path.relpath(subdir, self._logdir)
      runs_seen.add(run)
      if run not in self._directory_loaders:
        logger.info('- Adding run for relative directory %s', run)
        self._directory_loaders[run] = self._directory_loader_factory(subdir)
    stale_runs = set(self._directory_loaders) - runs_seen
    if stale_runs:
      for run in stale_runs:
        logger.info('- Removing run for relative directory %s', run)
        del self._directory_loaders[run]
    logger.info('Ending logdir traversal of %s', self._logdir)

  def get_run_events(self):
    """Returns tf.Event generators for each run's `DirectoryLoader`.

    Warning: the generators are stateful and consuming them will affect the
    results of any other existing generators for that run; calling code should
    ensure it takes events from only a single generator per run at a time.

    Returns:
      Dictionary containing an entry for each run, mapping the run name to a
      generator yielding tf.Event protobuf objects loaded from that run.
    """
    runs = list(self._directory_loaders)
    logger.info('Creating event loading generators for %d runs', len(runs))
    run_to_loader = collections.OrderedDict()
    for run_name in sorted(runs):
      loader = self._directory_loaders[run_name]
      run_to_loader[run_name] = self._wrap_loader_generator(loader.Load())
    return run_to_loader

  def _wrap_loader_generator(self, loader_generator):
    """Wraps `DirectoryLoader` generator to swallow `DirectoryDeletedError`."""
    try:
      for item in loader_generator:
        yield item
    except directory_watcher.DirectoryDeletedError:
      return
